/*
 * TimeTable Viewer
 * Copyright(c)1999 by Hiroaki Sengoku <sengoku@gcd.org>
 * Version 0.1	Jun 19, 1999
 * Version 0.2	Jun 20, 1999
 * Version 0.3	Jun 20, 1999
 * Version 0.4	Jun 21, 1999
 * Version 0.5	Jun 22, 1999	AboutDialog
 * Version 0.6	Jun 25, 1999	SelectTrain, HardButton, PrefsDialog
 * Version 0.7	Jun 26, 1999	scroll bug fix
 * Version 0.8	Jul  3, 1999	Arrive, ShareReg, History, Transfer, StopClock
 * Version 0.9	Jul  8,	1999	multi '>', ContFile, CommentMark, scramble
 * Version 1.0	Jul 31, 1999	split bar, DB, ;;CANCEL, '\' esc, HeadType, LRU
 * Version 1.1	Aug 26, 1999	moveRecord
 * Version 1.2	Aug 27, 2000	configFile, ;;PREC, DotPattern, Narrow
 * Version 1.3  Aug  1, 2001	GPL version
 */
#define TTV_DB
#include "Common.h"

/* List.c */
Boolean isKanji(Byte c);

/* Table.c */
extern short NItems;
extern short *ArriveTime;
extern short ArrivesColumn;
extern short OfsArrive;
extern short NarrowMode;

typedef struct {
    UInt renamedCategories;
    char categoryLabels[dmRecNumCategories][dmCategoryLength];
    Byte categoryUniqIDs[dmRecNumCategories];
    Byte lastUniqID;
    Byte reserved1;
    Word reserved2;
    Word reserved3;
    Byte sortOrder;
} MemoAppInfoType;

#define NComments	(26*2+1)

TimeTableType *CurrentTable;
char Title[STRMAX];
DmOpenRef MemoDB;
UInt TrainTimeCategory;
InsertPosType InsertPosMemo;
DmOpenRef TimeTableDB;
UInt TimeTableCategory;
VoidHand TimeTableAppInfoH;
InsertPosType InsertPosDB;
static char ContFileName[STRMAX];
short ContFileNo;
short NContFiles;
static long ContFileRecNo;
static VoidHand ContFileHandle;
static char *ContFileLastLine;
static char ContFileLastLineBuf[STRLINEMAX];
static char *ContFileNextLine;
static CommentMarkType CommentMark;
Word DotPattern[0x1F+1];
SByte NarrowList[NComments+1];

static char WeekChars[] =
"ALL ORD MON TUE WED THU FRI SAT SUN HOL "
"UR0 UR1 UR2 UR3 UR4 UR5 UR6 UR7 UR8 UR9 ";
WeekType Today;
ULong TodaySetSeconds;
static WeekType TodayDate;
long ConfigRecordNum;
static VoidHand ConfigHandle;

/* convert TrainTime DB into array of LineType */

void skipSpace(char **pp) {
    char *p;
    p = *pp;
    while (*p == ' ' || *p == '\t') p++;
    *pp = p;
}

static void skipLine(char **pp) {
    char *p, *q;
    p = *pp;
    q = StrChr(p,linefeedChr);
    if (q) {
	while (*q == linefeedChr) q++;
    } else {
	q = p + StrLen(p);
    }
    *pp = q;
}

static SByte conv2digit(char **pp) {
    char *p;
    short i;
    p = *pp;
    if ('0' <= *p && *p <= '9') {
	i = *p - '0';
	if ('0' <= *(p+1) && *(p+1) <= '9') {
	    i *= 10;
	    p++;
	    i += *p - '0';
	}
	p++;
	*pp = p;
    } else {
	i = -1;	/* error */
    }
    return i;
}

static short convDigits(char **pp) {
    char *p;
    short i = 0;
    Boolean err = true;
    p = *pp;
    while ('0' <= *p && *p <= '9') {
	err = false;
	i *= 10;
	i += *p - '0';
	p++;
    }
    if (err) {
	return -1;
    } else {
	*pp = p;
	return i;
    }
}

/* length of the first line and *pp points the next line */
short getLine(char **pp) {
    char *p, *q;
    short len;
    p = *pp;
    q = StrChr(p,linefeedChr);
    if (q) {
	len = q - p;
	q++;
    } else {
	len = StrLen(p);
	q = p + len;
    }
    *pp = q;
    return len;
}

/* Continuous File
1st file:	`;FileName' 1
		`;FileName'Real
2nd file:	`;FileName' 2
		body
`;FileName' must match */
Boolean isContFile(char **pp) {
    char *p, *q, *r;
    Boolean ret = false;
    short len, n, i;
    q = *pp;
    len = getLine(&q);
    p = *pp;
    for (i=len-1; i > 0; i--) {
	if (p[i] == ' ') {
	    r = &p[i+1];
	    n = convDigits(&r);
	    if (i > STRMAX-1) i = STRMAX - 1;
	    if (ContFileNo <= 0 && n == 1) {			/* 1st file */
		StrNCopy(ContFileName,p,i);
		ContFileName[i] = '\0';
		if (StrNCompare(q,ContFileName,i) == 0) {	/* 2nd line */
		    ContFileNo = n;
		    ret = true;
		}
	    } else if (0 < ContFileNo && n == ContFileNo + 1
		       && StrNCompare(p,ContFileName,i) == 0	/* 2nd file */
		       && ContFileName[i] == '\0') {
		ContFileNo = n;
		ret = true;
	    }
	    break;
	}
	if (p[i] < '0' || '9' < p[i]) break;	/* not digit */
    }
    if (ret) {
	*pp = q;
    } else {
	ContFileNo = -1;
	ContFileName[0] = '\0';
    }
    return ret;
}

static short commentChar(char c) {
    short ret;
    if ('A' <= c && c <= 'Z') {
	ret = c - 'A';
    } else if ('a' <= c && c <= 'z') {
	ret = c - 'a' + 26;
    } else if (c == '-') {
	ret = NoComment;	/* no comment */
    } else {
	ret = -1;	/* not comment char */
    }
    return ret;
}

static Boolean ishidden(SByte com, SByte *comments) {
    SByte *p, *q;
    if (NarrowMode <= 0) return false;	/* all */
    if (NarrowMode == 1) {
	if (com >= 0) return true;	/* without notes */
	return false;
    }
    p = NarrowList;
    while (*p >= 0) {
	if (*p == NoComment && com < 0) return false;
	if (com == *p) return false;
	q = comments;
	while (*q >= 0) {
	    if (*q == *p) return false;
	    q++;
	}
	p++;
    }
    return true;
}

static Boolean convItem(char **pp, ItemType *item,
			SByte *comments, short *cip,
			CommentMarkType *commentsMark) {
    char *p;
    short min;
    short com, c;
    CommentMarkType mark;
    short ci;
    p = *pp;
 again:
    mark = UnMark;
    ci = *cip;
    skipSpace(&p);
    com = commentChar(*p);
    if (com >= 0) {
	if (mark < commentsMark[com]) mark = commentsMark[com];
	p++;
	while ((c=commentChar(*p)) >= 0) {
	    if (commentsMark[c] >= Mark00000 && mark >= Mark00000) {
		mark = ((commentsMark[c] - Mark00000) | (mark - Mark00000))
		    + Mark00000;
	    } else if (mark < commentsMark[c]) mark = commentsMark[c];
	    if (ci < ExCommentsMax-1) comments[ci++] = c;
	    p++;
	} 
    }
    min = conv2digit(&p);
    if (min >= 0) {
	comments[ci++] = -1;	/* separater */
	if (ishidden(com,comments+*cip)) goto again;
	if (item) {
	    item->comment = com;
	    item->minute = min;
	    item->mark = (Byte)mark;
	}
    }
    skipSpace(&p);
    *pp = p;
    if (min >= 0) {
	*cip = ci;
	return true;
    }
    return false;
}

static void storeWord(Word w, Byte *p) {	/* *p = w */
    p[0] = (w >> 8);	/* MSB big endian */
    p[1] = (w & 0xff);	/* LSB */
}

static short convLine(char **pp, LineType *line,
		      short *cip, CommentMarkType *commentsMark) {
    static SByte comments[ExCommentsMax];	/* static to save stack area */
    SByte *cp;
    short n = 0;
    short i;
    *cip = 0;
    while (**pp && **pp != linefeedChr && n < NItems) {
	if (convItem(pp,(line ? &line->items[n] : NULL),
		     comments,cip,commentsMark)) n++;
	else {
	    skipLine(pp);
	    break;
	}
    }
    while (**pp == linefeedChr) (*pp)++;
    if (*cip == n) *cip = 0;	/* no extra comments */
    if (line) {
	storeWord((Word)*cip,&line->excom_msb);	/* line->excom = *cip; */
	cp = (SByte*)&line->items[n];
	for (i=0; i < *cip; i++) cp[i] = comments[i];
    }
    return n;
}

/* begin with '#' */
void titleLine(char **pp) {
    short length;
    char *p, *q;
    p = *pp;
    if (*p == '#' || *p == ';') p++;
    skipSpace(&p);
    q = p;
    length = getLine(&q);
    if (length >= STRMAX - 1) length = STRMAX - 1;
    StrNCopy(Title,p,length);
    Title[length] = '\0';
    *pp = q;
}

/* begin with [0-9]+: */
static short hourLine(char **pp, LineType *line, LineType **lines, short *lenP,
		      CommentMarkType *commentsMark) {
    short length = 0;
    short hour;
    short excom;
    short i, j;
    char *p;
    p = *pp;
    if (skipSpace(&p), (hour=conv2digit(&p)) >= 0 && *p++ == ':') {
    multi:
	i = convLine(&p,line,&excom,commentsMark);
	j = sizeof(LineType) + sizeof(ItemType) * i + sizeof(SByte) * excom;
	length += j;
	if (line) {
	    line->hour = hour;
	    line->length = i;
	    if (lines && lenP) lines[*lenP] = line;
	    (Byte*)line += j;
	}
	if (lenP) (*lenP)++;
	if (*p && *(p-1) != linefeedChr) goto multi;	/* multi line */
	*pp = p;
    } else {
	skipLine(pp);
    }
    return length;
}

/* begin with '[A-Za-z]:' */
static short commentLine(char **pp, char *buf, char **comments,
			 CommentMarkType *commentsMark,
			 char **narrows, short *narrowsLenP) {
    short length = 0;
    short comment;
    char *p, *q;
    SByte comms[NComments+1];
    short ncomms = 0;
    short i;
    p = *pp;
    while ((comment=commentChar(*p)) >= 0) {
	if (ncomms < NComments) comms[ncomms++] = comment;
	p++;
    }
    if (ncomms > 0 && *p++ == ':') {
	comment = comms[0];
	q = p;
	while (*q) {
	    if (*q == linefeedChr || *q == ';') break;
	    q++;
	}
	length = q - p;
	if (*q) skipLine(&q);
	if (buf) {
	    if (comments) comments[comment] = buf;
	    if (narrows && narrowsLenP) narrows[*narrowsLenP] = buf;
	    StrNCopy(buf,p,length);
	    buf[length] = '\0';
	}
	if (narrowsLenP) {
	    if ((*narrowsLenP) == NarrowMode) {
		for (i=0; i < ncomms; i++) {
		    NarrowList[i] = comms[i];
		}
		NarrowList[i] = -1;	/* delimitter */
	    }
	    (*narrowsLenP)++;
	}
	length++;	/* with delimitter */
	if (buf) {
	    for (i=0; i < ncomms; i++) {
		buf[length+i] = comms[i];
	    }
	    buf[length+ncomms] = -1;	/* delimitter */
	}
	length += (ncomms + 1);
	if (commentsMark) commentsMark[comment] = CommentMark;
	*pp = q;
	return length;
    } else {
	skipLine(pp);
	return 0;
    }
}

/* begin with '$' */
static short arriveLine(char **pp, char *buf, char **arrives,
			short *lenP) {
    short stn = 0;
    short length = 0;
    char *p;
    short comment;
    short min;
    Boolean eod = false;
    char trans[STRLINEMAX];
    short trlen = 0;
    short trpos = 0;
    short i;
    Boolean kanji;
    p = *pp;
    p++;
    while (*p && *p != ':') {
	if (buf) buf[length] = *p;
	length++;
	p++;
    }
    if (buf) {
	if (arrives && lenP) arrives[*lenP] = buf;
	buf[length] = '\0';
    }
    if (lenP) {
	stn = *lenP;
	(*lenP)++;
    }
    length++;
    trlen = 0;
    trpos = length;
    length += 2;
    if (buf) {
	buf[trpos] = 0;
	buf[trpos+1] = 0;
    }
    skipSpace(&p);
    while (*p && *p != linefeedChr) {
	if (*p == ';') {
	    skipLine(&p);
	    break;
	}
	if (*p == '>') {	/* transfer */
	    p++;
	    skipSpace(&p);
	    kanji = false;
	    while(*p && *p != linefeedChr && *p != ' '
		  && *p != '+' && trlen < STRLINEMAX-2) {
		if (!kanji && *p == '\\') p++;	/* escape char */
		kanji = isKanji(*p);
		trans[trlen++] = *p++;
	    }
	    trans[trlen++] = '\0';
	    skipSpace(&p);
	    if (*p == '+') {
		p++;
		skipSpace(&p);
		trans[trlen++] = convDigits(&p);
		skipSpace(&p);
	    } else {
		trans[trlen++] = 0;
	    }
	    continue;
	}
	eod = false;
	comment = commentChar(*p);
	if (comment < 0) {
	    if (('0' <= *p && *p <= '9') || *p == '=') {
		if (buf) buf[length] = (char)-1;
		length++;
		if (*p == '=') {
		    min = -2;
		    p++;
		} else {
		    min = convDigits(&p);
		}
		if (buf) storeWord((Word)min,(Byte*)&buf[length]);
		length += 2;
		eod = true;
	    } else {
		p++;
	    }
	    skipSpace(&p);
	    continue;
	}
	if (buf) buf[length] = (char)comment;
	length++;
	p++;
    }
    if (!eod) {
	if (buf) {
	    min = -1;
	    buf[length] = (char)-1;
	    storeWord((Word)min,(Byte*)&buf[length+1]);
	}
	length += 3;
    }
    if (buf) buf[length] = -2;	/* end of arriveLine */
    length++;
    if (trlen) {
	i = length - trpos;
	if (buf) storeWord((Word)i,(Byte*)&buf[trpos]);
	for (i=0; i < trlen; i++) {
	    if (buf) buf[length] = trans[i];
	    length++;
	}
	if (buf) buf[length] = '\0';
	length++;
    }
    if (*p == linefeedChr) p++;
    *pp = p;
    return length;
}

static WeekType weekTypeToken(char **pp) {
    char str[4];
    char *p, *q;
    short i = 0;
    WeekType ret = Errday;
    p = *pp;
    while (*p != '[' && *p != '*' && *p && *p != linefeedChr) {
	p++;
    }
    if (*p == '[' || *p == '*') {
	p++;
    } else {
	*pp = p;
	return ret;
    }
    while(*p != ']' && *p != '*' && *p != linefeedChr && *p) {
	if (i < 3) str[i++] = *p;
	p++;
    }
    if (*p == ']') p++;
    str[i] = '\0';
    *pp = p;
    q = StrStr(WeekChars,str);
    if (q) {
	ret = (q - WeekChars)/4 + Allday;
    }
    return ret;
}

/* begin with '[' */
static Boolean weekTypeLine(char **pp) {
    WeekType week;
    Boolean ret = false;
    Boolean comb;
    char *p;
    p = *pp;
    while (*p && *p != linefeedChr) {
	comb = (*p == '*');
	if (!comb && ret) break;
	week = weekTypeToken(&p);
	if (week == Errday) {
	    continue;
	} else if (week == Allday) {
	    ret = true;
	} else if (week == Weekday) {
	    if ((Monday <= Today && Today <= Friday)
		|| Today == Weekday) ret = true;
	} else if (week == TodayDate) {
	    ret = true;
	} else if (Today == Weekday) {
	    if ((Monday <= week && week <= Friday)
		|| week == Weekday) ret = true;
	} else if (Today == week) {
	    ret = true;
	} else if (comb) {
	    ret = false;	/* conbination with previous conditions */
	}
    }
    skipLine(&p);
    *pp = p;
    return ret;
}

/* convert binary string to integer */
static long btoi(char **pp) {
    long ret = 0;
    char *p;
    p = *pp;
    if (*p != '0' && *p != '1') return -1;
    do {
	ret <<= 1;
	ret |= (*p - '0');
	p++;
    } while (*p == '0' || *p == '1');
    *pp = p;
    return ret;
}

/* begin with ';;' */
static void directiveLine(char **pp) {
    char *p;
    short len, n;
    long dot;
    p = *pp;
    len = getLine(pp);
    if (len <= 0) return;
    if (StrNCompare(p,"UNMARK",len) == 0) {
	CommentMark = UnMark;
    } else if (StrNCompare(p,"SUCC",len) == 0) {
	CommentMark = SuccMark;
    } else if (StrNCompare(p,"PREC",len) == 0) {
	CommentMark = PrecMark;
    } else if (StrNCompare(p,"MIDDLE",len) == 0) {
	CommentMark = MiddleMark;
    } else if (StrNCompare(p,"MARK",len) == 0) {
	CommentMark = Mark;
    } else if (StrNCompare(p,"DOTMARK",len) == 0) {
	CommentMark = DotMark;
    } else if (StrNCompare(p,"DDOTMARK",len) == 0) {
	CommentMark = DDotMark;
    } else if (StrNCompare(p,"CANCEL",len) == 0) {
	CommentMark = CancelMark;
    } else if (StrNCompare(p,"MARK",4) == 0
	       && (p += 4, (n=btoi(&p)) >= 0) && n <= 0x1F) {
	CommentMark = Mark00000 + n;
	if (*p++ == '=' && (dot=btoi(&p)) >= 0) {
	    DotPattern[n] = (Word)dot;
	}
    }
}

static char contNextChar(char **pp) {
    char *p;
    short index, n;
    p = *pp;
    if (ContFileNo == 0) {	/* initialize */
	NContFiles = 0;
	ContFileHandle = NULL;
	ContFileLastLine = NULL;
	index = 0;
	ContFileLastLineBuf[index] = '\0';
    } else {
	if (ContFileNextLine) {
	    if (*p == '\0') {	/* get from ContFileLastLineBuf */
		p = *pp = ContFileNextLine;
		ContFileNextLine = NULL;
	    }
	    if (*p) return *p;
	}
	if (ContFileNo < 0) {	/* no more continue */
	    if (*p == '\0') {
		if (ContFileHandle) MemHandleUnlock(ContFileHandle);
		ContFileHandle = NULL;
		return '\0';
	    }
	    return *p;
	}
	if (p < ContFileLastLine) return *p;
	for (index=0; index < STRLINEMAX-1 && *p; index++)
	    ContFileLastLineBuf[index] = *p++;
	ContFileLastLineBuf[index] = '\0';
    }
    if (ContFileHandle) MemHandleUnlock(ContFileHandle);
    ContFileNextLine = NULL;
    if (ContFileRecNo >= TimeTableRecordNumMin && TimeTableDB) {
	UInt recordNum = ContFileRecNo - TimeTableRecordNumMin;
	ContFileHandle = DmQueryNextInCategory(TimeTableDB,&recordNum,
					       TimeTableCategory);
	ContFileRecNo = recordNum + TimeTableRecordNumMin;
    } else {
	UInt recordNum = ContFileRecNo;
	ContFileHandle = DmQueryNextInCategory(MemoDB,&recordNum,
					       TrainTimeCategory);
	ContFileRecNo = recordNum;
    }
    ContFileRecNo++;
    n = ContFileNo;
    if (ContFileHandle) {
	p = MemHandleLock(ContFileHandle);
	ContFileLastLine = p + MemHandleSize(ContFileHandle) - 1;
	while (*ContFileLastLine != linefeedChr) ContFileLastLine--;
	ContFileLastLine++;
	if (isContFile(&p)) {
	    NContFiles++;
	    for (; index < STRLINEMAX-1 && *p && *p != linefeedChr; index++)
		ContFileLastLineBuf[index] = *p++;
	    if (*p == linefeedChr) ContFileLastLineBuf[index++] = *p++;
	    ContFileLastLineBuf[index] = '\0';
	    if (*p == linefeedChr) p++;
	    ContFileNextLine = p;
	    p = *pp = ContFileLastLineBuf;
	    return *p;
	} else if (n <= 0) {	/* not continuous file */
	    NContFiles = 1;
	    *pp = p;
	    return *p;
	} else {		/* end of continuous file */
	    MemHandleUnlock(ContFileHandle);
	    ContFileHandle = NULL;
	}
    }
    ContFileNo = -1;
    p = *pp = ContFileLastLineBuf;
    return *p;
}

static void passDB(long recordNum, TimeTableType *table,
		   short *tlenp, short *clenp, short *alenp,
		   CommentMarkType *commentsMark) {
    static short tcount;
    short count;
    char *p;
    short tlen, clen, alen;
    Boolean ok = true;
#ifdef PROFILE
    HostFPrintF(HostFp,"passDB(long recordNum,...) %lu\n",TimGetTicks());
#endif
    ContFileNo = 0;
    ContFileRecNo = recordNum;
    contNextChar(&p);
    skipSpace(&p);
    titleLine(&p);
    table->LinesLen = 0;
    table->ArrivesLen = 0;
    table->NarrowsLen = NarrowsOffset;
    NarrowList[0] = -1;
    tlen = clen = alen = 0;
    count = 0;
    if (table->LinesBuf == NULL) tcount = count;	/* pass 1 */
    CommentMark = Mark;	/* underline */
    while (contNextChar(&p)) {
	count++;
	skipSpace(&p);
	if (*p == '[') {
	    ok = weekTypeLine(&p);
	    if (ok && table->LinesBuf == NULL) {	/* pass 1 */
		tcount = count;		/* reload from here in pass 2 */
		tlen = 0;
		table->LinesLen = 0;
	    }
	    continue;
	}
	if (!ok) {
	    skipLine(&p);
	    continue;
	}
	if (*p == ';') {
	    if (*(p+1) == ';') {
		p += 2;
		directiveLine(&p);
	    } else {
		skipLine(&p);	/* comment line */
	    }
	} else if (commentChar(*p) >= 0) {
	    clen += commentLine(&p,(table->CommentsBuf ?
				    table->CommentsBuf + clen : NULL),
				table->Comments,commentsMark,
				table->Narrows,&table->NarrowsLen);
	} else if (*p == '#') {
	    titleLine(&p);	/* title */
	} else if ('0' <= *p && *p <= '9') {
	    if (table->LinesBuf == NULL
		|| count >= tcount)	/* in pass 2, reload from tcount */
		tlen += hourLine(&p,(LineType*)(table->LinesBuf ?
						table->LinesBuf + tlen : NULL),
				 table->Lines,&table->LinesLen,
				 commentsMark);
	    else skipLine(&p);
	} else if (*p == '$') {
	    alen += arriveLine(&p,(table->ArrivesBuf ?
				   table->ArrivesBuf + alen : NULL),
			       table->Arrives,&table->ArrivesLen);
	} else {
	    skipLine(&p);
	}
    }
    *tlenp = tlen;
    *clenp = clen;
    *alenp = alen;
}

static PrivErrType convTrainTimeDB(long recordNum, TimeTableType *table) {
    CommentMarkType commentsMark[NComments];
    short tlen, clen, alen;
    short tlen2, clen2, alen2;
    short i, j;
    for (i=0; i < NComments; i++) commentsMark[i] = UnMark;
    for (i=0; i <= 0x1F; i++) {	/* initialize DotPattern */
	DotPattern[i] = 0;
	for (j=0x10; j > 0; j >>= 1) {
	    DotPattern[i] <<= 2;
	    if (i & j) DotPattern[i] |= 0x3;
	}
	DotPattern[i] <<= 2;
    }
    passDB(recordNum,table,&tlen,&clen,&alen,commentsMark);	/* pass 1 */
    if (ContFileHandle) {
	PrivErrDisplay("Can't happen: ContFileHandle != NULL",true);
	MemHandleUnlock(ContFileHandle);
	ContFileHandle = NULL;
    }
    if (tlen > 0 && table->LinesLen > 0) {
	if (table->LinesBuf) MemPtrFree(table->LinesBuf);
	table->LinesBuf = MemPtrNew(tlen);
	if (!table->LinesBuf) {
	    PrivErrExtraStr = "table->LinesBuf";
	    return privErrNotEnoughSpace;
	}
	if (table->Lines) MemPtrFree(table->Lines);
	table->Lines = MemPtrNew(sizeof(LineType*) * table->LinesLen);
	if (!table->Lines) {
	    PrivErrExtraStr = "table->Lines";
	    return privErrNotEnoughSpace;
	}
    }
    if (clen > 0) {
	if (table->Comments) MemPtrFree(table->Comments);
	table->Comments = MemPtrNew(sizeof(char*) * NComments);
	if (!table->Comments) {
	    PrivErrExtraStr = "table->Comments";
	    return privErrNotEnoughSpace;
	}
	if (table->CommentsBuf) MemPtrFree(table->CommentsBuf);
	table->CommentsBuf = MemPtrNew(clen);
	if (!table->CommentsBuf) {
	    PrivErrExtraStr = "table->CommentsBuf";
	    return privErrNotEnoughSpace;
	}
	for (i=0; i < NComments; i++) table->Comments[i] = "";
    }
    if (table->Narrows) MemPtrFree(table->Narrows);
    table->Narrows = MemPtrNew(sizeof(char*) * table->NarrowsLen);
    if (!table->Narrows) {
	PrivErrExtraStr = "table->Narrows";
	return privErrNotEnoughSpace;
    }
    if (alen > 0 && table->ArrivesLen > 0) {
	if (table->ArrivesBuf) MemPtrFree(table->ArrivesBuf);
	table->ArrivesBuf = MemPtrNew(alen);
	if (!table->ArrivesBuf) {
	    PrivErrExtraStr = "table->ArrivesBuf";
	    return privErrNotEnoughSpace;
	}
	if (table->Arrives) MemPtrFree(table->Arrives);
	table->Arrives = MemPtrNew(sizeof(char*) * table->ArrivesLen);
	if (!table->Arrives) {
	    PrivErrExtraStr = "table->Arrives";
	    return privErrNotEnoughSpace;
	}
    }
    passDB(recordNum,table,&tlen2,&clen2,&alen2,commentsMark);	/* pass 2 */
    if (ContFileHandle) {
	PrivErrDisplay("Can't happen: ContFileHandle != NULL",true);
	MemHandleUnlock(ContFileHandle);
	ContFileHandle = NULL;
    }
    if (tlen2 != tlen) {
	PrivErrExtraStr = "tlen2 != tlen";
    } else if (clen2 != clen) {
	PrivErrExtraStr = "clen2 != clen";
    } else if (alen2 != alen) {
	PrivErrExtraStr = "alen2 != alen";
    }
    if (PrivErrExtraStr) return privErrCanNotHappen;
    if (tlen <= 0 || table->LinesLen <= 0) table->LinesLen = 0;
#if 0
    tlen = 0;
    for (i=0; i < table->LinesLen; i++) tlen += table->Lines[i]->length;
    if (tlen <= 0) table->LinesLen = 0;
#endif
    if (alen > 0 && table->ArrivesLen > 0) {
	if (ArriveTime) MemPtrFree(ArriveTime);
	ArriveTime = MemPtrNew(sizeof(short) * table->ArrivesLen
			       * ArrivesColumn);
	if (!ArriveTime) {
	    PrivErrExtraStr = "ArriveTime";
	    return privErrNotEnoughSpace;
	}
	if (OfsArrive >= table->ArrivesLen) OfsArrive = 0;
    } else {
	table->ArrivesLen = 0;
    }
#ifdef DEBUG
    for (i=0; i <= 0x1F; i++) {
	HostFPrintF(HostFp,"DotPattern[%2x]=",i);
	for (j=0x2000; j > 0; j >>= 1) {
	    if (DotPattern[i] & j) HostFPrintF(HostFp,"1");
	    else HostFPrintF(HostFp,"0");
	}
	HostFPrintF(HostFp,"\n");
    }
#endif
    return privErrNone;
}

TimeTableType *newTimeTable(void) {
    TimeTableType *table = MemPtrNew(sizeof(TimeTableType));
    if (!table) return NULL;
    table->recordNum = -1;
    table->Lines = NULL;
    table->LinesLen = 0;
    table->LinesBuf = NULL;
    table->Comments = NULL;
    table->CommentsBuf = NULL;
    table->Arrives = NULL;
    table->ArrivesLen = 0;
    table->ArrivesBuf = NULL;
    table->NarrowsLen = 0;
    table->Narrows = NULL;
    return table;
}

void disposeTimeTable(TimeTableType *table) {
    if (table->Lines) MemPtrFree(table->Lines);
    if (table->LinesBuf) MemPtrFree(table->LinesBuf);
    if (table->Comments) MemPtrFree(table->Comments);
    if (table->CommentsBuf) MemPtrFree(table->CommentsBuf);
    if (table->Arrives) MemPtrFree(table->Arrives);
    if (table->ArrivesBuf) MemPtrFree(table->ArrivesBuf);
    if (table->Narrows) MemPtrFree(table->Narrows);
    MemPtrFree(table);
}

static short dateToken(char **pp) {
    short mon, day;
    skipSpace(pp);
    mon = (short)conv2digit(pp);
    skipSpace(pp);
    if (mon < 0 || **pp != '/') return -1;
    (*pp)++;
    skipSpace(pp);
    day = (short)conv2digit(pp);
    skipSpace(pp);
    if (day < 0) return -1;
    return day + (mon << 8);
}

static WeekType parseConfig(VoidHand h) {
    WeekType week = Errday;
    short from, end, today;
    DateTimeType dateTime;
    ULong now = TimGetSeconds();
    char *p;
    if (!h) return week;
    TimSecondsToDateTime(now,&dateTime);
    if (dateTime.hour <= MidNight) {
	now -= ((dateTime.hour+1) * 60L * 60L);	/* yesterday */
	TimSecondsToDateTime(now,&dateTime);
    }
    today = dateTime.day + (dateTime.month << 8);
#ifdef DEBUG
    HostFPrintF(HostFp,"parseConfig: today = %x\n",today);
#endif
    p = MemHandleLock(h);
    skipLine(&p);
    while (*p) {
	skipSpace(&p);
	if (*p == '[') {
	    week = weekTypeToken(&p);
	    skipLine(&p);
	} else if (*p == ';') {
	    skipLine(&p);
	} else if ('0' <= *p && *p <= '9') {
	    from = dateToken(&p);
	    if (*p == '-') {
		p++;
		end = dateToken(&p);
	    } else {
		end = from;
	    }
#ifdef DEBUG
	    HostFPrintF(HostFp,"parseConfig: week = %d, date = %x-%x\n",
			week,from,end);
#endif
	    if (0 <= from && from <= today
		&& 0 <= end && today <= end) goto exit;
	    skipLine(&p);
	} else {
	    skipLine(&p);
	}
    }
    week = Errday;
 exit:
    MemHandleUnlock(h);
    return week;
}

static Boolean loadDB(long recordNum) {
    TimeTableType *table;
    PrivErrType Err;
#ifdef PROFILE
    HostFPrintF(HostFp,"loadDB(long recordNum) %lu\n",TimGetTicks());
#endif
    setToday(false);
    table = newTimeTable();
    if (table) {
	ULong diff = TimGetSeconds() - TodaySetSeconds;
	TodayDate = Errday;
	if (diff >= 10 * 60	/* 10 min */
	    && ConfigHandle)
	    TodayDate = parseConfig(ConfigHandle);
	Err = convTrainTimeDB(recordNum,table);
    } else {
	PrivErrExtraStr = "newTimeTable";
	Err = privErrNotEnoughSpace;
    }
    if (Err) {
	PrivErrDisplay(PrivErrString(Err),true);
    } else if (table->LinesLen > 0) {
	if (CurrentTable) disposeTimeTable(CurrentTable);
	table->recordNum = recordNum;
	CurrentTable = table;
	return true;
    }
    disposeTimeTable(table);
    return false;
}

Boolean checkTitle(long recordNum, char *title) {
    VoidHand record;
    char str[STRMAX];
    char *p;
    short i;
    if (recordNum >= TimeTableRecordNumMin && TimeTableDB) {
	record = DmQueryRecord(TimeTableDB,recordNum - TimeTableRecordNumMin);
    } else {
	record = DmQueryRecord(MemoDB,recordNum);
    }
    if (record == NULL) return false;
    p = MemHandleLock(record);
    ContFileNo = 0;
    (void)isContFile(&p);
    for (i=0; i < STRMAX-1; i++) {
	if (!*p || *p == linefeedChr) break;
	str[i] = *p++;
    }
    str[i] = '\0';
    MemHandleUnlock(record);
    return StrStr(str,title) != NULL;
}

long searchFile(char *p) {
    UInt recordNum;
    long ret;
    VoidHand record;
    if (TimeTableDB) {
	recordNum = 0;
	while ((record=DmQueryNextInCategory(TimeTableDB,&recordNum,
					     TimeTableCategory))) {
	    ret = recordNum++;
	    if (checkTitle(ret+TimeTableRecordNumMin,p))
		return ret + TimeTableRecordNumMin;
	}
    }
    recordNum = 0;
    while ((record=DmQueryNextInCategory(MemoDB,&recordNum,
					 TrainTimeCategory))) {
	ret = recordNum++;
	if (checkTitle(ret,p)) return ret;
    }
    return -1;
}

long searchTransfer(char *p, short *ofs) {
    static char *files = NULL;
    static long recordNum;
    UInt num;
    long ret;
    VoidHand record;
    short i;
    if (p) {
	files = p;
	if (TimeTableDB) recordNum = TimeTableRecordNumMin;
	else recordNum = 0;
    } else if (files == NULL) return -1;
    while (*files) {
	i = StrLen(files);
	*ofs = files[i+1];
	if (recordNum >= TimeTableRecordNumMin) {
	    num = recordNum - TimeTableRecordNumMin;
	    while ((record=DmQueryNextInCategory(TimeTableDB,&num,
						 TimeTableCategory))) {
		ret = num++;
		recordNum = num + TimeTableRecordNumMin;
		if (checkTitle(ret+TimeTableRecordNumMin,files))
		    return ret + TimeTableRecordNumMin;
	    }
	    recordNum = 0;
	}
	num = recordNum;
	while ((record=DmQueryNextInCategory(MemoDB,&num,
					     TrainTimeCategory))) {
	    ret = num++;
	    recordNum = num;
	    if (checkTitle(ret,files)) return ret;
	}
	files += i + 2;	/* next filename */
	if (TimeTableDB) recordNum = TimeTableRecordNumMin;
	else recordNum = 0;
    }
    files = NULL;
    return -1;	/* never reach */
}

static Int compareRecords(char *r1, char *r2, Int other,
			  SortRecordInfoPtr info1, SortRecordInfoPtr info2,
			  VoidHand appInfoH) {
    return StrCompare(r1,r2);
}

static long insertPosition(DmOpenRef db, UInt category, VoidHand srcH) {
    long ret = -1;
    UInt recordNum;
    VoidHand dstH;
    char *src;
    Int cmp;
    src = MemHandleLock(srcH);
    recordNum = 0;
    while ((dstH=DmQueryNextInCategory(db,&recordNum,category))) {
	cmp = StrCompare(src,MemHandleLock(dstH));
	MemHandleUnlock(dstH);
	if (cmp <= 0) break;
	recordNum++;
    }
    MemHandleUnlock(srcH);
    ret = recordNum;
    return ret;
}

static long moveRecordSub(DmOpenRef srcDB, UInt src, UInt scat,
			  DmOpenRef dstDB, UInt dst, UInt dcat,
			  short nconts, Byte sortOrder) {
    VoidHand srcH, dstH;
    Byte *srcP, *dstP;
    UInt at = dst;
    ULong size;
    UInt attr;
    Err err;
    long ret = -1;
    short i;
    for (i=0; i < nconts; i++) {
	srcH = DmQueryNextInCategory(srcDB,&src,scat);
	if (!srcH) return -1;
	size = MemHandleSize(srcH);
	if (i == 0) {
	    if (sortOrder) {
		long pos = insertPosition(dstDB,dcat,srcH);
		if (pos >= 0) at = pos;
	    }
	    ret = at;
	}
	dstH = DmNewRecord(dstDB,&at,size);
	if (!dstH) return -1;
#ifdef DEBUG
	HostFPrintF(HostFp,"moveRecordSub: %d %d -> %d\n",i,src,at);
#endif
	srcP = MemHandleLock(srcH);
	dstP = MemHandleLock(dstH);
	DmWrite(dstP,0,srcP,size);
	MemHandleUnlock(srcH);
	MemHandleUnlock(dstH);
	DmReleaseRecord(dstDB,at,true);
	if (dcat != dmAllCategories && dcat != dmUnfiledCategory) {
	    err = DmRecordInfo(dstDB,at,&attr,NULL,NULL);
	    if (err) {
		PrivErrSetCode(err);
		PrivErrDisplay("moveRecordSub: Can't get info of record",
			       false);
		attr = 0;
	    }
	    attr |= dcat;
	    err = DmSetRecordInfo(dstDB,at,&attr,NULL);
	    if (err) {
		PrivErrSetCode(err);
		PrivErrDisplay("moveRecordSub: Can't set info of record",
			       false);
	    }
	}
	DmDeleteRecord(srcDB,src);
	at++;
    }
    return ret;
}

long moveRecord(long src, short nconts) {
    DmOpenRef ttvDB = NULL;
    DmOpenRef memoDB = NULL;
    long ret = -1;
    ttvDB = DmOpenDatabaseByTypeCreator('DATA',APPID,dmModeReadWrite);
    if (!ttvDB) {
	PrivErrSetCode(DmGetLastErr());
	PrivErrDisplay("Can't open TimeTable DB",false);
	goto close;
    }
    memoDB = DmOpenDatabaseByTypeCreator('DATA','memo',dmModeReadWrite);
    if (!memoDB) {
	PrivErrSetCode(DmGetLastErr());
	PrivErrDisplay("Can't open MemoPad",false);
	goto close;
    }
    if (src >= TimeTableRecordNumMin) {		/* TTV -> MemoPad */
	src -= TimeTableRecordNumMin;
	ret = moveRecordSub(ttvDB,src,TimeTableCategory,
			    memoDB,(InsertPosMemo == InsertPosTop ? 0 : -1),
			    TrainTimeCategory,
			    nconts,InsertPosMemo == InsertPosAlphabetical);
    } else {					/* MemoPad -> TTV */
	ret = moveRecordSub(memoDB,src,TrainTimeCategory,
			    ttvDB,(InsertPosDB == InsertPosTop ? 0 : -1),
			    TimeTableCategory,
			    nconts,InsertPosDB == InsertPosAlphabetical);
	if (ret >= 0) ret += TimeTableRecordNumMin;
    }
 close:
    if (ttvDB) DmCloseDatabase(ttvDB);
    if (memoDB) DmCloseDatabase(memoDB);
    return ret;
}

void moveAllRecords(long index) {
    DmOpenRef ttvDB = NULL;
    DmOpenRef memoDB = NULL;
    UInt recordNum;
    VoidHand record;
    ttvDB = DmOpenDatabaseByTypeCreator('DATA',APPID,dmModeReadWrite);
    if (!ttvDB) {
	PrivErrSetCode(DmGetLastErr());
	PrivErrDisplay("Can't open TimeTable DB",false);
	goto close;
    }
    memoDB = DmOpenDatabaseByTypeCreator('DATA','memo',dmModeReadWrite);
    if (!memoDB) {
	PrivErrSetCode(DmGetLastErr());
	PrivErrDisplay("Can't open MemoPad",false);
	goto close;
    }
    recordNum = 0;
    if (index >= TimeTableRecordNumMin) {	/* TTV -> MemoPad */
	while ((record=DmQueryNextInCategory(ttvDB,&recordNum,
					     TimeTableCategory))) {
	    if (moveRecordSub(ttvDB,recordNum,TimeTableCategory,
			      memoDB,(InsertPosMemo == InsertPosTop ? 0 : -1),
			      TrainTimeCategory,1,0) < 0) break;
	    recordNum++;
	}
	if (InsertPosMemo == InsertPosAlphabetical)
	    DmInsertionSort(memoDB,(DmComparF*)&compareRecords,0);
    } else {					/* MemoPad -> TTV */
	while ((record=DmQueryNextInCategory(memoDB,&recordNum,
					     TrainTimeCategory))) {
	    if (moveRecordSub(memoDB,recordNum,TrainTimeCategory,
			      ttvDB,(InsertPosDB == InsertPosTop ? 0 : -1),
			      TimeTableCategory,1,0) < 0) break;
	    recordNum++;
	}
	if (InsertPosDB == InsertPosAlphabetical)
	    DmInsertionSort(ttvDB,(DmComparF*)&compareRecords,0);
    }
 close:
    if (ttvDB) DmCloseDatabase(ttvDB);
    if (memoDB) DmCloseDatabase(memoDB);
}

void sortAllRecords(long index) {
    DmOpenRef db = NULL;
    if (index >= TimeTableRecordNumMin) {	/* TTV */
	db = DmOpenDatabaseByTypeCreator('DATA',APPID,dmModeReadWrite);
	if (!db) {
	    PrivErrSetCode(DmGetLastErr());
	    PrivErrDisplay("Can't open TimeTable DB to sort",false);
	    return;
	}
    } else {					/* MemoPad */
	db = DmOpenDatabaseByTypeCreator('DATA','memo',dmModeReadWrite);
	if (!db) {
	    PrivErrSetCode(DmGetLastErr());
	    PrivErrDisplay("Can't open MemoPad to sort",false);
	    return;
	}
    }
    DmInsertionSort(db,(DmComparF*)&compareRecords,0);
    if (db) DmCloseDatabase(db);
}

void deleteDupRecords(long index) {
    DmOpenRef db = NULL;
    UInt cat;
    UInt prevNum, nextNum;
    VoidHand prevH, nextH;
    char *prevP, *nextP;
    short i;
    Boolean same;
    if (index >= TimeTableRecordNumMin) {	/* TTV */
	db = DmOpenDatabaseByTypeCreator('DATA',APPID,dmModeReadWrite);
	if (!db) {
	    PrivErrSetCode(DmGetLastErr());
	    PrivErrDisplay("Can't open TimeTable DB to delete dups",false);
	    return;
	}
	cat = TimeTableCategory;
    } else {					/* MemoPad */
	db = DmOpenDatabaseByTypeCreator('DATA','memo',dmModeReadWrite);
	if (!db) {
	    PrivErrSetCode(DmGetLastErr());
	    PrivErrDisplay("Can't open MemoPad to delete dups",false);
	    return;
	}
	cat = TrainTimeCategory;
    }
    prevNum = 0;
    prevH = DmQueryNextInCategory(db,&prevNum,cat);
    if (!prevH) goto close;
    prevP = MemHandleLock(prevH);
    nextNum = prevNum + 1;
    while ((nextH=DmQueryNextInCategory(db,&nextNum,cat))) {
	nextP = MemHandleLock(nextH);
	i = 0;
	while (prevP[i] && nextP[i] && prevP[i] == nextP[i]) i++;
	same = (prevP[i] == nextP[i]);
	MemHandleUnlock(prevH);
#ifdef DEBUG
	HostFPrintF(HostFp,"compare %d %d -> %d(%d)\n",
		    prevNum,nextNum,same,i);
#endif
	if (same) DmDeleteRecord(db,prevNum);
	prevNum = nextNum;
	prevH = nextH;
	prevP = nextP;
	nextNum = prevNum + 1;
    }
    MemHandleUnlock(prevH);
 close:
    if (db) DmCloseDatabase(db);
}

static short isConfigFile(UInt recordNum) {
    char *p;
    short type = 0;	/* assume not config file */
    VoidHand h = DmQueryNextInCategory(MemoDB,&recordNum,dmAllCategories);
    if (!h) return -1;	/* not exist */
    p = MemHandleLock(h);
    if (StrNCompare(p,"TrainTime",9) == 0
	&& p[9] == linefeedChr) {
	type = 1;	/* TrainTime ConfigFile */
    } else if (StrNCompare(p,"TimeTable Viewer",16) == 0
	       && p[16] == linefeedChr) {
	type = 2;	/* TimeTable Viewer ConfigFile */
    }
    MemHandleUnlock(h);
    ConfigHandle = h;
    return type;
}

static void configFile(void) {
    UInt recordNum;
    UInt trainNum = 0;
    VoidHand trainH = NULL;
    short type;
#ifdef PROFILE
    HostFPrintF(HostFp,"configFile(void) %lu\n",TimGetTicks());
#endif
    recordNum = 0;
    while ((type=isConfigFile(recordNum)) >= 0) {
	switch (type) {
	case 1:
	    trainNum = recordNum;
	    trainH = ConfigHandle;
	    break;
	case 2:
	    ConfigRecordNum = recordNum;
	    ConfigHandle = ConfigHandle;
	    return;
	}
	recordNum++;
    }
    if (trainH) {
	ConfigRecordNum = trainNum;
	ConfigHandle = trainH;
	return;
    }
    ConfigRecordNum = -1;
    ConfigHandle = NULL;
}

void initializeDB(void) {
    VoidHand appInfoH;
#ifdef PROFILE
    HostFPrintF(HostFp,"initializeDB(void) %lu\n",TimGetTicks());
#endif
    MemoDB = DmOpenDatabaseByTypeCreator('DATA','memo',dmModeReadOnly);
    if (MemoDB) {
	TrainTimeCategory = CategoryFind(MemoDB,"TrainTime");
	appInfoH = getAppInfo(MemoDB);
	if (appInfoH) {
	    MemoAppInfoType *appInfoP;
	    appInfoP = MemHandleLock(appInfoH);
	    if (appInfoP->sortOrder) InsertPosMemo = InsertPosAlphabetical;
	    MemHandleUnlock(appInfoH);
	}
    } else {
	PrivErrSetCode(DmGetLastErr());
	PrivErrDisplay("Can't open MemoPad",true);
    }
}

void initializeDBAfterPrefs(void) {
    if (ConfigRecordNum < 0 || isConfigFile(ConfigRecordNum) <= 0)
	configFile();
}
