/*
 * TimeTable Viewer
 * Copyright(c)1999 by Hiroaki Sengoku <sengoku@gcd.org>
 * Version 0.1	Jun 19, 1999
 * Version 0.2	Jun 20, 1999
 * Version 0.3	Jun 20, 1999
 * Version 0.4	Jun 21, 1999
 * Version 0.5	Jun 22, 1999	AboutDialog
 * Version 0.6	Jun 25, 1999	SelectTrain, HardButton, PrefsDialog
 * Version 0.7	Jun 26, 1999	scroll bug fix
 * Version 0.8	Jul  3, 1999	Arrive, ShareReg, History, Transfer, StopClock
 * Version 0.9	Jul  8,	1999	multi '>', ContFile, CommentMark, scramble
 * Version 1.0	Jul 31, 1999	split bar, DB, ;;CANCEL, '\' esc, HeadType, LRU
 * Version 1.1	Aug 26, 1999	vertical split bar
 * Version 1.2	Aug 27, 2000	SeparatorMap
 * Version 1.3  Aug  1, 2001	GPL version
 */
#include "Common.h"

/* Main.c */
extern MenuBarPtr CurrentMenuBar;

/* Table.c */
extern Boolean setTimeTable(long recordNum);
extern FileMenuType *getFileMenuItem(ULong now, short item);
extern void setFileMenuItem(ULong now, short item,
			    char *str, short len, long recordNum);
extern long TrainTimeRecordNum;

/* DB.c */
extern void initializeDB(void);
extern void initializeDBAfterPrefs(void);
extern Boolean isContFile(char **pp);
extern void skipSpace(char **pp);
extern short getLine(char **pp);
extern void moveAllRecords(long index);
extern void sortAllRecords(long index);
extern void deleteDupRecords(long index);
extern short ContFileNo;
extern DmOpenRef MemoDB;
extern UInt TrainTimeCategory;
extern DmOpenRef TimeTableDB;
extern UInt TimeTableCategory;
extern InsertPosType InsertPosDB;
extern InsertPosType InsertPosMemo;

/* Dialog.c */
extern void stepAlarmTriggered(SysAlarmTriggeredParamType *cmdPBP,
			       ULong recID, LocalID dbID, UInt cardNo);
extern Byte SeparatorMap[];
extern Boolean SeparatorDirty;

/* TimeTable List */
FormPtr ListFormPtr;
MenuBarPtr ListMenuBar;

short ListX;	/* [pixel] */
short ListY;
short ListWidth;
#define BarMargin	2
#define BarWidth	7
#define ScrollBarMargin	(BarMargin+BarWidth+BarMargin)
#define SepMargin	3
Boolean StnBarFlag;
Boolean ListBarFlag;
short StnSepWidth;
short StnWidth;
short RowHeight;
short NRows;	/* [row] */
short StnTop;
short ListTop;
short handleListPenRow;

StnIndexType *StnTrains;
char *StnNameBuf;
short NStnTrains;
short SelStn;
long *AllTrains;
short NAllTrains;
long *Trains;
short NTrains;
ListOrderType ListOrder;

/* TimeTableList Form */

static VoidHand queryTimeTable(long recordNum) {
    VoidHand ret;
    if (recordNum >= TimeTableRecordNumMin && TimeTableDB) {
	ret = DmQueryRecord(TimeTableDB,recordNum-TimeTableRecordNumMin);
    } else {
	ret = DmQueryRecord(MemoDB,recordNum);
    }
    return ret;
}

static void addMenuItem(long recordNum) {
    ListPtr list;
    short item;
    VoidHand memo;
    char *p;
    short width, len;
    Boolean ok;
    list = FrmGetObjectPtr(ListFormPtr,FrmGetObjectIndex
			   (ListFormPtr,TableListAddPopUpList));
    item = LstGetSelection(list) - 1;
    if (item >= 0) {
	memo = queryTimeTable(recordNum);
	if (memo) {
	    char title[MENU_LEN_MAX];
	    p = MemHandleLock(memo);
	    ContFileNo = 0;
	    (void)isContFile(&p);
	    skipSpace(&p);
	    if (*p == ';') p++;
	    skipSpace(&p);
	    StrNCopy(title,p,MENU_LEN_MAX-1);
	    MemHandleUnlock(memo);
	    p = StrChr(title,linefeedChr);
	    if (p) *p = '\0'; else title[MENU_LEN_MAX-1] = '\0';
	    len = StrLen(title);
	    width = MenuItemWidth;
	    ok = true;
	    FntCharsInWidth(title,&width,&len,&ok);
	    setFileMenuItem(TimGetSeconds(),item,title,len,recordNum);
	}
    }
}

static void rmMenuItem(void) {
    ListPtr list;
    short item;
    ControlPtr control;
    list = FrmGetObjectPtr(ListFormPtr,FrmGetObjectIndex
			   (ListFormPtr,TableListAddPopUpList));
    item = LstGetSelection(list) - 1;
    if (item >= 0) setFileMenuItem(TimGetSeconds(),item,"",0,-1);
    control = FrmGetObjectPtr(ListFormPtr,FrmGetObjectIndex
			      (ListFormPtr,TableListRemove));
    CtlHideControl(control);
}

static void invertStn(short stn) {
    RectangleType r;
    if (stn < StnTop || NStnTrains <= stn
	|| StnTop+NRows <= stn) return;
    r.topLeft.x = ListX;
    r.topLeft.y = ListY + RowHeight * (stn - StnTop);
    r.extent.x = StnSepWidth;
    r.extent.y = RowHeight;
    WinInvertRectangle(&r,0);
}

static void drawStnListLine(short i) {
    short len, width;
    char *p;
    Boolean ok;
    RectangleType r;
    r.topLeft.x = ListX;
    r.extent.x = StnSepWidth;
    r.topLeft.y = ListY + RowHeight*i;
    r.extent.y = RowHeight;
    WinEraseRectangle(&r,0);
    if (StnTop+i >= NStnTrains) return;
    width = StnSepWidth;
    p = StnTrains[StnTop+i].name;
    len = StrLen(p);
    ok = true;
    FntCharsInWidth(p,&width,&len,&ok);
    WinDrawChars(p,len,r.topLeft.x,r.topLeft.y);
    if (StnTop+i == SelStn) invertStn(SelStn);
}

static Boolean isSeparator(Byte c) {
    if (c < ' ') return true;
    c -= ' ';
    if ((c >> 3) >= SeparatorMapLen) return false;
    return (SeparatorMap[c >> 3] & (0x80 >> (c & 0x7))) != 0;
}

Boolean isKanji(Byte c) {
    return (0x80 <= c && c <= 0x9f) || (0xe0 <= c && c <= 0xff);
}

static void drawListLine(short i) {
    VoidHand memo;
    short len, width;
    char *p, *q;
    Boolean ok;
    RectangleType r;
    r.topLeft.y = ListY + RowHeight*i;
    r.extent.y = RowHeight;
    if (StnSepWidth > 0) {	/* 2 column mode */
	width = StnSepWidth + SepMargin;
	if (StnBarFlag) width += ScrollBarMargin;
	r.topLeft.x = ListX + width;
	r.extent.x = ListWidth - width;
	if (ListBarFlag) r.extent.x -= BarWidth + BarMargin;
    } else {
	r.topLeft.x = ListX;
	r.extent.x = ListWidth;
	if (ListBarFlag) r.extent.x -= BarWidth + BarMargin;
    }
    WinEraseRectangle(&r,0);
    if (ListTop+i >= NTrains) return;
    memo = queryTimeTable(Trains[ListTop+i]);
    if (memo) {
	p = MemHandleLock(memo);
	ContFileNo = 0;
	(void)isContFile(&p);
	skipSpace(&p);
	if (*p == ';') p++;
	skipSpace(&p);
	if (StnSepWidth > 0) {	/* 2 column mode */
	    while (isSeparator(*p)) p++;
	    while (*p && !isSeparator(*p)) {
		if (isKanji(*p)) p++;
		p++;
	    }
	    while (isSeparator(*p)) p++;
	}
	q = p;
	len = getLine(&q);
	ok = true;
	width = r.extent.x;
	FntCharsInWidth(p,&width,&len,&ok);
	WinDrawChars(p,len,r.topLeft.x,r.topLeft.y);
	MemHandleUnlock(memo);
    }
}

static void updateStnListBar(Boolean draw) {
    ScrollBarPtr bar;
    bar = FrmGetObjectPtr(ListFormPtr,FrmGetObjectIndex
			  (ListFormPtr,StnListScrollBar));
    if (NStnTrains <= NRows) {
	StnBarFlag = false;
	if (bar) {
	    bar->attr.usable = false;
	}
    } else {
	if (!StnBarFlag) {
	    RectangleType r;
	    r.topLeft.x = ListX + StnSepWidth;
	    r.topLeft.y = ListY;
	    r.extent.x = ScrollBarMargin + SepMargin;
	    r.extent.y = RowHeight * NRows;
	    WinEraseRectangle(&r,0);
	}
	StnBarFlag = true;
	if (bar) {
	    bar->attr.usable = true;
	    if (draw) {
		bar->bounds.topLeft.x = ListX + StnSepWidth + BarMargin;
		bar->bounds.extent.y = RowHeight * NRows;
		SclSetScrollBar(bar,StnTop,0,NStnTrains-NRows,NRows);
		SclDrawScrollBar(bar);
	    } else {
		SclSetScrollBar(bar,StnTop,0,NStnTrains-NRows,NRows);
	    }
	}
    }
}

static void updateListBar(Boolean draw) {
    ScrollBarPtr bar;
    bar = FrmGetObjectPtr(ListFormPtr,FrmGetObjectIndex
			  (ListFormPtr,TableListScrollBar));
    if (NTrains <= NRows) {
	if (ListBarFlag) {
	    RectangleType r;
	    r.topLeft.x = ListX + ListWidth - BarWidth;
	    r.topLeft.y = ListY;
	    r.extent.x = BarWidth;
	    r.extent.y = RowHeight * NRows;
	    WinEraseRectangle(&r,0);
	}
	ListBarFlag = false;
	if (bar) {
	    bar->attr.usable = false;
	}
    } else {
	if (!ListBarFlag) {
	    RectangleType r;
	    r.topLeft.x = ListX + ListWidth - (BarWidth + BarMargin);
	    r.topLeft.y = ListY;
	    r.extent.x = BarWidth + BarMargin;
	    r.extent.y = RowHeight * NRows;
	    WinEraseRectangle(&r,0);
	}
	ListBarFlag = true;
	if (bar) {
	    bar->attr.usable = true;
	    if (draw) {
		bar->bounds.extent.y = RowHeight * NRows;
		SclSetScrollBar(bar,ListTop,0,NTrains-NRows,NRows);
		SclDrawScrollBar(bar);
	    } else {
		SclSetScrollBar(bar,ListTop,0,NTrains-NRows,NRows);
	    }
	}
    }
}

static void scrollList(short lines) {
    RectangleType r, vacated;
    short dist, from, end, i;
    DirectionType dir;
    if (lines == 0 || NTrains <= NRows) return;
    if (lines > 0) {
	if (ListTop + NRows + lines > NTrains) {
	    lines = NTrains - NRows - ListTop;
	    if (lines == 0) return;
	}
	dir = up;
	dist = lines;
	from = NRows - 1 - lines;
	end = NRows - 1;
    } else {
	if (ListTop + lines < 0) {
	    lines = - ListTop;
	    if (lines == 0) return;
	}
	dir = down;
	dist = -lines;
	from = 0;
	end = -lines - 1;
    }
    r.topLeft.y = ListY;
    r.extent.y = RowHeight * NRows;
    if (StnSepWidth > 0) {
	r.topLeft.x = ListX + StnSepWidth + SepMargin;
	r.extent.x = ListWidth - StnSepWidth - SepMargin;
	if (StnBarFlag) {
	    r.topLeft.x += ScrollBarMargin;
	    r.extent.x -= ScrollBarMargin;
	}
	if (ListBarFlag) {
	    r.extent.x -= BarWidth + BarMargin;
	}
    } else {
	r.topLeft.x = ListX;
	r.extent.x = ListWidth - (BarWidth + BarMargin);
    }
    if (dist < NRows) {
	WinScrollRectangle(&r,dir,dist*RowHeight,&vacated);
	WinEraseRectangle(&vacated,0);
    } else {
	WinEraseRectangle(&r,0);
	from = 0;
	end = NRows - 1;
    }
    ListTop += lines;
    for (i=from; i <= end; i++) drawListLine(i);
    updateListBar(false);
}

static void scrollStnList(short lines) {
    RectangleType r, vacated;
    short dist, from, end, i;
    DirectionType dir;
    if (lines == 0 || NStnTrains <= NRows) return;
    if (lines > 0) {
	if (StnTop + NRows + lines > NStnTrains) {
	    lines = NStnTrains - NRows - StnTop;
	    if (lines == 0) return;
	}
	dir = up;
	dist = lines;
	from = NRows - 1 - lines;
	end = NRows - 1;
    } else {
	if (StnTop + lines < 0) {
	    lines = - StnTop;
	    if (lines == 0) return;
	}
	dir = down;
	dist = -lines;
	from = 0;
	end = -lines - 1;
    }
    r.topLeft.y = ListY;
    r.extent.y = RowHeight * NRows;
    r.topLeft.x = ListX;
    r.extent.x = StnSepWidth;
    if (dist < NRows) {
	WinScrollRectangle(&r,dir,dist*RowHeight,&vacated);
	WinEraseRectangle(&vacated,0);
    } else {
	WinEraseRectangle(&r,0);
	from = 0;
	end = NRows - 1;
    }
    StnTop += lines;
    for (i=from; i <= end; i++) drawStnListLine(i);
    updateStnListBar(false);
}

static void drawStnList(void) {
    short i;
    updateStnListBar(true);
    for (i=0; i < NRows; i++) drawStnListLine(i);
}

static void drawList(void) {
    short splitX, splitY;
    short i;
#ifdef PROFILE
    HostFPrintF(HostFp,"drawList(void) %lu\n",TimGetTicks());
#endif
    splitX = ListX + StnSepWidth;
    if (splitX <= 0) splitX = 1;
    splitY = ListY + RowHeight*NRows;
    WinDrawLine(splitX,splitY,splitX,splitY+BarWidth);
    updateListBar(true);
    for (i=0; i < NRows; i++) drawListLine(i);
}

static void selectStn(short stn, Boolean draw) {
    Trains = &AllTrains[StnTrains[stn].itop];
    NTrains = StnTrains[stn].ilen;
    if (stn != SelStn) {
	ListTop = 0;
	if (draw) {
	    if (SelStn >= 0) invertStn(SelStn);
	    if (stn >= 0) invertStn(stn);
	    drawList();
	}
    }
    SelStn = stn;
}

static Boolean handleListPenSplit(EventPtr e) {
    static short px;
    short ex = e->screenX;
    short x;
    short width = 3;
    RectangleType r;
    x = ListX + StnSepWidth;
    if (x <= 0) x = 1;
    r.extent.x = width;
    r.topLeft.y = ListY;
    r.extent.y = RowHeight * NRows;
    if (ex <= 0) ex = 1;
    else if (ex > ListY + ListWidth - 50) ex = ListY + ListWidth - 50;
    if (e->eType == penDownEvent) {
	px = x;
	r.topLeft.x = px - 1;
	WinInvertRectangle(&r,0);
    } else if (e->eType == penMoveEvent && ex != px) {
	r.topLeft.x = ex - 1;
	if (ex <= px - r.extent.x || px + r.extent.x <= ex) {
	    WinInvertRectangle(&r,0);
	    r.topLeft.x = px - 1;
	    WinInvertRectangle(&r,0);
	} else if (ex < px) {
	    r.extent.x = px - ex;
	    WinInvertRectangle(&r,0);
	    r.topLeft.x += width;
	    WinInvertRectangle(&r,0);
	} else if (px < ex) {
	    r.extent.x = ex - px;
	    r.topLeft.x += width - r.extent.x;
	    WinInvertRectangle(&r,0);
	    r.topLeft.x -= width;
	    WinInvertRectangle(&r,0);
	}
	px = ex;
    } else if (e->eType == penUpEvent && px >= 0) {
	short splitX, splitY;
	r.topLeft.x = px - 1;
	WinInvertRectangle(&r,0);
	splitX = ListX + StnSepWidth;
	if (splitX <= 0) splitX = 1;
	splitY = ListY + RowHeight*NRows;
	WinEraseLine(splitX,splitY,splitX,splitY+BarWidth);
	width = StnSepWidth;	/* save original */
	StnSepWidth = px - ListX;
	if (StnSepWidth < 10) StnSepWidth = 0;
	r.topLeft.x++;
	r.extent.x = SepMargin;
	if (StnBarFlag) r.extent.x += ScrollBarMargin;
	WinEraseRectangle(&r,0);
	if (StnSepWidth > 0) {
	    if (width <= 0) {
		ListTop = 0;
		selectStn(SelStn,false);
	    }
	    drawStnList();
	} else {
	    if (width > 0) {
		if (SelStn >= 0) ListTop = StnTrains[SelStn].itop;
		else ListTop = 0;
	    }
	    Trains = AllTrains;
	    NTrains = NAllTrains;
	}
	drawList();
	handleListPenRow = -1;
    }
    return true;
}

static Boolean tapScroll(EventPtr e) {
    static ULong lastTick;
    ULong now;
    short ey = e->screenY;
    Boolean handled = false;
    now = TimGetTicks();
    if (now - lastTick < 10) return false;	/* 0.1 sec */
    lastTick = now;
    if (ey < ListY) {
	scrollStnList(-1);
	selectStn(StnTop,true);
	handled = true;
    } else if (ListY+RowHeight*NRows <= ey
	       && StnTop+NRows < NStnTrains) {
	scrollStnList(1);
	selectStn(StnTop+NRows-1,true);
	handled = true;
    }
    return handled;
}

static Boolean handleListPen(EventPtr e) {
    static RectangleType tr;
    static Boolean on;
    short ex, ey;
    short row;
    short listL, listR;
    short splitX, splitY;
    RectangleType r;
    Boolean stn = false;
    Boolean handled = false;
    ex = e->screenX;
    ey = e->screenY;
    splitX = ListX + StnSepWidth;
    if (splitX <= 0) splitX = 1;
    splitY = ListY + RowHeight*NRows;
    if (handleListPenRow == -2) {
	return handleListPenSplit(e);	/* split bar */
    } else if (e->eType == penDownEvent
	       && splitX-2 <= ex && ex <= splitX+2
	       && splitY <= ey && ey <= splitY + BarWidth) {
	handleListPenRow = -2;
	return handleListPenSplit(e);
    }
    if (StnSepWidth > 0) {
	listL = ListX + StnSepWidth + SepMargin;
	if (StnBarFlag) listL += ScrollBarMargin;
    } else {
	listL = ListX;
    }
    listR = ListX + ListWidth;
    if (ListBarFlag) listR -= BarWidth + BarMargin;
    if (ey < ListY || ListY+RowHeight*NRows <= ey
	|| ex < ListX || listR <= ex
	|| (ListX+StnSepWidth <= ex && ex < listL)) {
	row = -1;
    } else {
	row = (ey - ListY) / RowHeight;
	r.topLeft.y = ListY + RowHeight * row;
	r.extent.y = RowHeight;
	if (ex < ListX+StnSepWidth) {
	    stn = true;
	    r.topLeft.x = ListX;
	    r.extent.x = StnSepWidth;
	} else {
	    r.topLeft.x = listL;
	    r.extent.x = listR - listL;
	}
    }
    if (stn) {
	if (StnTop+row >= NStnTrains) row = -1;
    } else {
	if (ListTop+row >= NTrains) row = -1;
    }
    switch (e->eType) {
    case penDownEvent:
	if (row >= 0) {
	    if (stn) {
		handleListPenRow = -3;
		selectStn(StnTop+row,true);
		handled = true;
	    } else {
		handleListPenRow = row;
		WinInvertRectangle(&r,0);
		tr = r;
		on = true;
		handled = true;
	    }
	} else {
	    handleListPenRow = -1;
	    r.topLeft.x = r.topLeft.y = 0;
	    r.extent.x = r.extent.y = 0;
	    on = false;
	}
	break;
    case penMoveEvent:
	if (handleListPenRow >= 0) {
	    if ((on && row != handleListPenRow)
		|| (!on && row == handleListPenRow)) {
		WinInvertRectangle(&tr,0);
		on = !on;
	    }
	    handled = true;
	} else if (stn && row >= 0) {
	    selectStn(StnTop+row,true);
	    handled = true;
	} else if (handleListPenRow == -3) {
	    handled = tapScroll(e);
	}
	break;
    case penUpEvent:
	if (handleListPenRow >= 0) {
	    if (row >= 0 && row == handleListPenRow
		&& setTimeTable(Trains[ListTop+row])) {
		addMenuItem(Trains[ListTop+row]);
		FrmGotoForm(MainForm);
		handled = true;
	    }
	    if (on) WinInvertRectangle(&tr,0);
	}
	handleListPenRow = -1;
	break;
    default:
    }
    return handled;
}

static short getStnName(char **pp, char *stn) {
    short len = 0;
    Byte *p;
    p = *pp;
    while (*p == ' ' || *p == ';' || isSeparator(*p)) p++;
    while (*p && !isSeparator(*p) && len < STRMAX) {
	if (isKanji(*p)) {
	    if (stn) stn[len] = *p;
	    len++;
	    p++;
	}
	if (stn) stn[len] = *p;
	len++;
	p++;
    }
    if (stn) stn[len] = '\0';
    return len;
}

static char *nextRecord(DmOpenRef db, UInt *recordNumP, UInt cat,
			VoidHand *hP, UInt *nextNumP) {
    VoidHand record;
    char *top, *p;
    ContFileNo = 0;
    record = DmQueryNextInCategory(db,recordNumP,cat);
    if (!record) return NULL;
    top = MemHandleLock(record);
    *hP = record;
    *nextNumP = *recordNumP;
    if (isContFile(&top)) {	/* skip cont files */
	(*nextNumP)++;
	while((record=DmQueryNextInCategory(db,nextNumP,cat))) {
	    p = MemHandleLock(record);
	    if (!isContFile(&p)) {
		MemHandleUnlock(record);
		break;
	    }
	    MemHandleUnlock(record);
	    (*nextNumP)++;
	}
    } else {
	(*nextNumP)++;
    }
    return top;
}

static ULong getULong(BytePtr p) {
    ULong ret = 0;
    if (!p) return 0;
    ret = *p++;
    ret = ((ret << 8) | *p++);
    ret = ((ret << 8) | *p++);
    ret = ((ret << 8) | *p);
    return ret;
}

static void loadTrainTimeIndexMix(short *tlenP, long *tindex,
				  short *slenP, StnIndexType *sindex,
				  short *blenP, char *buf) {
    short tlen = *tlenP;
    short slen = *slenP;
    short blen = *blenP;
    char *ttvP, *memoP, *p;
    VoidHand ttvH, memoH;
    UInt ttvNum, ttvNext, memoNum, memoNext;
    ULong ttvText, memoText;
    char stn[STRMAX+1], pstn[STRMAX+1];
    short len;
    pstn[0] = '\0';
    ttvNum = memoNum = 0;
    ttvP = nextRecord(TimeTableDB,&ttvNum,TimeTableCategory,&ttvH,&ttvNext);
    ttvText = getULong(ttvP);
    memoP = nextRecord(MemoDB,&memoNum,TrainTimeCategory,&memoH,&memoNext);
    memoText = getULong(memoP);
    while (ttvP || memoP) {
	if (!memoP) goto ttvFirst;
	if (!ttvP) goto memoFirst;
	if (ttvText < memoText) goto ttvFirst;
	if (ttvText > memoText) goto memoFirst;
	if (StrNCompare(ttvP,memoP,STRMAX) > 0) goto memoFirst;
    ttvFirst:
	p = ttvP;
	len = getStnName(&p,stn);
	if (tindex) tindex[tlen] = ttvNum + TimeTableRecordNumMin;
	tlen++;
	MemHandleUnlock(ttvH);
	ttvNum = ttvNext;
	ttvP = nextRecord(TimeTableDB,&ttvNum,TimeTableCategory,
			  &ttvH,&ttvNext);
	ttvText = getULong(ttvP);
	goto common;
    memoFirst:
	p = memoP;
	len = getStnName(&p,stn);
	if (tindex) tindex[tlen] = memoNum;
	tlen++;
	MemHandleUnlock(memoH);
	memoNum = memoNext;
	memoP = nextRecord(MemoDB,&memoNum,TrainTimeCategory,
			   &memoH,&memoNext);
	memoText = getULong(memoP);
    common:
	if (StrCompare(stn,pstn) != 0) {	/* new station */
	    if (sindex) {
		sindex[slen].itop = tlen-1;
		sindex[slen].name = &buf[blen];
		if (slen > 0)
		    sindex[slen-1].ilen = tlen-1 - sindex[slen-1].itop;
	    }
	    slen++;
	    if (buf) {
		StrNCopy(&buf[blen],stn,len);
		buf[blen+len] = '\0';
	    }
	    blen += len + 1;
	    StrCopy(pstn,stn);
	}
    }
    if (sindex) {
	if (slen > 0) sindex[slen-1].ilen = tlen - sindex[slen-1].itop;
    }
    *slenP = slen;
    *blenP = blen;
    *tlenP = tlen;
}

static void loadTrainTimeIndexSub(short *tlenP, long *tindex,
				  DmOpenRef db, UInt cat, long ofs,
				  short *slenP, StnIndexType *sindex,
				  short *blenP, char *buf) {
    char *p;
    VoidHand h;
    short tlen = *tlenP;
    short slen = *slenP;
    short blen = *blenP;
    UInt recordNum = 0;
    UInt nextNum;
    char stn[STRMAX+1], pstn[STRMAX+1];
    short len;
    pstn[0] = '\0';
    while((p=nextRecord(db,&recordNum,cat,&h,&nextNum))) {
	len = getStnName(&p,stn);
	MemHandleUnlock(h);
	if (StrCompare(stn,pstn) != 0) {	/* new station */
	    if (sindex) {
		sindex[slen].itop = tlen;
		sindex[slen].name = &buf[blen];
		if (slen > 0)
		    sindex[slen-1].ilen = tlen - sindex[slen-1].itop;
	    }
	    slen++;
	    if (buf) {
		StrNCopy(&buf[blen],stn,len);
		buf[blen+len] = '\0';
	    }
	    blen += len + 1;
	    StrCopy(pstn,stn);
	}
	if (tindex) tindex[tlen] = recordNum + ofs;
	tlen++;
	recordNum = nextNum;
    }
    if (sindex) {
	if (slen > 0) sindex[slen-1].ilen = tlen - sindex[slen-1].itop;
    }
    *slenP = slen;
    *blenP = blen;
    *tlenP = tlen;
}

static void loadTrainTimeIndex(short *tlenP, long *tindex,
			       short *slenP, StnIndexType *sindex,
			       short *blenP, char *buf) {
    *tlenP = *slenP = *blenP = 0;
    if (!TimeTableDB || ListOrder == ListOrder_Memo_Only) {
	loadTrainTimeIndexSub(tlenP,tindex,MemoDB,TrainTimeCategory,0,
			      slenP,sindex,blenP,buf);
    } else if (ListOrder == ListOrder_DB_Only) {
	loadTrainTimeIndexSub(tlenP,tindex,TimeTableDB,
			      TimeTableCategory,TimeTableRecordNumMin,
			      slenP,sindex,blenP,buf);
    } else if (ListOrder == ListOrder_DB_Memo) {
	loadTrainTimeIndexSub(tlenP,tindex,TimeTableDB,
			      TimeTableCategory,TimeTableRecordNumMin,
			      slenP,sindex,blenP,buf);
	loadTrainTimeIndexSub(tlenP,tindex,MemoDB,TrainTimeCategory,0,
			      slenP,sindex,blenP,buf);
    } else if (ListOrder == ListOrder_Memo_DB) {
	loadTrainTimeIndexSub(tlenP,tindex,MemoDB,TrainTimeCategory,0,
			      slenP,sindex,blenP,buf);
	loadTrainTimeIndexSub(tlenP,tindex,TimeTableDB,
			      TimeTableCategory,TimeTableRecordNumMin,
			      slenP,sindex,blenP,buf);
    } else {
	loadTrainTimeIndexMix(tlenP,tindex,
			      slenP,sindex,blenP,buf);
    }
}

static void setListControl(void) {
    short i;
    ControlPtr control;
    FileMenuType *menu = getFileMenuItem(TimGetSeconds(),0);
    if (TrainTimeRecordNum < 0) {
	control	= FrmGetObjectPtr(ListFormPtr,FrmGetObjectIndex
				  (ListFormPtr,TableListCancel));
	CtlHideControl(control);
    }
    for (i=0; i < MENU_NITEMS; i++) {
	if (menu[i].recordNum < 0) {
	    setPopup(ListFormPtr,TableListAddPopUp,TableListAddPopUpList,i+1);
	    break;
	}
    }
    control = FrmGetObjectPtr(ListFormPtr,FrmGetObjectIndex
			      (ListFormPtr,TableListRemove));
    CtlHideControl(control);
}

static Boolean openList(void) {
    RectangleType r;
    short i, j, k;
    short stnBufLen;
#ifdef PROFILE
    HostFPrintF(HostFp,"openList(void) %lu\n",TimGetTicks());
#endif
    if (!ListFormPtr) ListFormPtr = FrmGetFormPtr(TableListForm);
    if (!ListFormPtr) return false;
    FntSetFont(stdFont);
    FrmGetObjectBounds(ListFormPtr,FrmGetObjectIndex
		       (ListFormPtr,TableListTable),&r);
    ListX = r.topLeft.x;
    ListY = r.topLeft.y;
    ListWidth = r.extent.x;
    RowHeight = FntLineHeight();
    NRows = r.extent.y / RowHeight;
    loadTrainTimeIndex(&NAllTrains,NULL,	/* pass 1 */
		       &NStnTrains,NULL,&stnBufLen,NULL);
    if (AllTrains) MemPtrFree(AllTrains);
    if (NAllTrains > 0) {
	AllTrains = MemPtrNew(NAllTrains*sizeof(long));
	if (!AllTrains) {
	    PrivErrExtraStr = "AllTrains";
	    PrivErrDisplay(PrivErrString(privErrNotEnoughSpace),true);
	    goto error;
	}
    } else {
	AllTrains = NULL;
    }
    if (StnTrains) MemPtrFree(StnTrains);
    if (NStnTrains > 0) {
	StnTrains = MemPtrNew(NStnTrains*sizeof(StnIndexType));
	if (!StnTrains) {
	    PrivErrExtraStr = "StnTrains";
	    PrivErrDisplay(PrivErrString(privErrNotEnoughSpace),true);
	    goto error;
	}
    } else {
	StnTrains = NULL;
    }
    if (StnNameBuf) MemPtrFree(StnNameBuf);
    if (stnBufLen > 0) {
	StnNameBuf = MemPtrNew(stnBufLen);
	if (!StnNameBuf) {
	    PrivErrExtraStr = "StnNameBuf";
	    PrivErrDisplay(PrivErrString(privErrNotEnoughSpace),true);
	    goto error;
	}
    } else {
	StnNameBuf = NULL;
    }
    loadTrainTimeIndex(&i,AllTrains,	/* pass 2 */
		       &j,StnTrains,&k,StnNameBuf);
    if (NAllTrains != i) {
	PrivErrExtraStr = "NAllTrains != i";
	PrivErrDisplay(PrivErrString(privErrCanNotHappen),true);
	goto error;
    }
    if (NStnTrains != j) PrivErrExtraStr = "NStnTrains != j";
    else if (stnBufLen != k) PrivErrExtraStr = "stnBufLen != k";
    if (PrivErrExtraStr) {
	PrivErrDisplay(PrivErrString(privErrCanNotHappen),true);
	goto error;
    }
    if (StnTop + NRows > NStnTrains) {
	StnTop = NStnTrains - NRows;
	if (StnTop < 0) StnTop = 0;
    }
    if (StnSepWidth > 0) {	/* 2 column mode */
	selectStn(SelStn,false);
    } else {			/* 1 column */
	Trains = AllTrains;
	NTrains = NAllTrains;
    }
    if (ListTop + NRows > NTrains) {
	ListTop = NTrains - NRows;
	if (ListTop < 0) ListTop = 0;
    }
#ifdef DEBUG
    for (i=0; i < NStnTrains; i++) {
	HostFPrintF(HostFp,"%s: itop %d, ilen %d\n",
		    StnTrains[i].name,StnTrains[i].itop,StnTrains[i].ilen);
    }
#endif
    handleListPenRow = -1;
    CurrentMenuBar = ListMenuBar;
    return true;
 error:
    FrmDeleteForm(ListFormPtr);
    ListFormPtr = NULL;
    return false;
}

static void closeList(void) {
    if (StnTrains) {
	MemPtrFree(StnTrains);
	StnTrains = NULL;
    }
    if (StnNameBuf) {
	MemPtrFree(StnNameBuf);
	StnNameBuf = NULL;
    }
    if (AllTrains) {
	MemPtrFree(AllTrains);
	AllTrains = NULL;
    }
    Trains = NULL;
}

/* Handles Events */

Boolean TableListFormHandleEvent(EventPtr e) {
    Boolean handled = false;
    
    CALLBACK_PROLOGUE

    switch (e->eType) {
    case frmOpenEvent:
	if (openList()) {
	    FrmDrawForm(ListFormPtr);
	    setListControl();
	    if (StnSepWidth > 0) drawStnList();	/* 2 column mode */
	    drawList();
	    handled = true;
	}
	break;
    case frmCloseEvent:
	closeList();
	CurrentMenuBar = NULL;
    /*	handled = false;  Let the system to delete the current form */
	ListFormPtr = NULL;
	break;
    case frmUpdateEvent:
	FrmDrawForm(ListFormPtr);
	setListControl();
	if (StnSepWidth > 0) drawStnList();	/* 2 column mode */
	drawList();
	handled = true;
	break;
    case menuEvent:
	MenuEraseStatus(NULL);
	switch(e->data.menu.itemID) {
	case OptionsMenuAbout:
	    doAboutDialog();
	    break;
	case OptionsMenuDispPrefs:
	    doDispPrefsDialog();
	    break;
	case OptionsMenuListPrefs:
	    {
		ListOrderType listorder = ListOrder;
		InsertPosType insertposdb = InsertPosDB;
		InsertPosType insertposmemo = InsertPosMemo;
		Boolean dirty = false;
		doListPrefsDialog();
		if (ListOrder != listorder) dirty = true;
		if (InsertPosDB == InsertPosAlphabetical
		    && InsertPosDB != insertposdb) {
		    sortAllRecords(TimeTableRecordNumMin);
		    dirty = true;
		}
		if (InsertPosMemo == InsertPosAlphabetical
		    && InsertPosMemo != insertposmemo) {
		    sortAllRecords(0);
		    dirty = true;
		}
		if (dirty || SeparatorDirty) goto reopen;
	    }
	    break;
	case OptionsMenuButtonPrefs:
	    doButtonPrefsDialog();
	    break;
	case ListMenuMoveToTTV:
	    moveAllRecords(0);
	    goto reopen;
	case ListMenuMoveToMemo:
	    moveAllRecords(TimeTableRecordNumMin);
	reopen:
	    closeList();
	    if (openList()) {
		setListControl();
		if (StnSepWidth > 0) drawStnList();	/* 2 column mode */
		drawList();
	    }
	    break;
	case ListMenuDeleteDupsTTV:
	    deleteDupRecords(TimeTableRecordNumMin);
	    goto reopen;
	case ListMenuDeleteDupsMemo:
	    deleteDupRecords(0);
	    goto reopen;
	}
	break;
    case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	case TableListCancel:
	    if (TrainTimeRecordNum >= 0
		&& setTimeTable(TrainTimeRecordNum)) {
		handled = true;
		FrmGotoForm(MainForm);
	    }
	    break;
	case TableListRemove:
	    rmMenuItem();
	    handled = true;
	    break;
	}
	break;
    case popSelectEvent:
	if (e->data.popSelect.listID == TableListAddPopUpList) {
	    short item = e->data.popSelect.selection;
	    ControlPtr control
		= FrmGetObjectPtr(ListFormPtr,FrmGetObjectIndex
				  (ListFormPtr,TableListRemove));
	    item--;
	    if (item < 0) {
		CtlHideControl(control);
	    } else if (item < MENU_NITEMS) {
		FileMenuType *menu = getFileMenuItem(TimGetSeconds(),item);
		if (menu->recordNum < 0) {
		    CtlHideControl(control);
		} else {
		    CtlShowControl(control);
		}
	    }
	}
	break;
    case penDownEvent:
    case penMoveEvent:
    case penUpEvent:
	handled = handleListPen(e);
	break;
    case keyDownEvent:
	if (ChrIsHardKey(e->data.keyDown.chr)) {
	    e->eType = appStopEvent;	/* leave main event loop */
	    break;
	}
	if (StnSepWidth > 0) {
	    short stn = SelStn;
	    if (e->data.keyDown.chr == pageUpChr && stn > 0) {
		stn--;
	    focus:
		if (stn < StnTop) {
		    scrollStnList(stn - StnTop);
		} else if (StnTop+NRows <= stn) {
		    scrollStnList(stn-(StnTop+NRows-1));
		}
		selectStn(stn,true);
		handled = true;
	    } else if (e->data.keyDown.chr == pageDownChr
		       && stn < NStnTrains-1) {
		stn++;
		goto focus;
	    }
	} else {
	    if (e->data.keyDown.chr == pageUpChr) {
		scrollList(-1);
		handled = true;
	    } else if (e->data.keyDown.chr == pageDownChr) {
		scrollList(1);
		handled = true;
	    }
	}
	break;
    case sclRepeatEvent:
	if (e->data.sclRepeat.scrollBarID == StnListScrollBar) {
	    scrollStnList(e->data.sclRepeat.newValue
			  - e->data.sclRepeat.value);
	} else {
	    scrollList(e->data.sclRepeat.newValue - e->data.sclRepeat.value);
	}
	break;
    case nilEvent:
	if (handleListPenRow == -3 && e->penDown) tapScroll(e);
	drawClock();
	break;
    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

void initializeList(void) {
    VoidHand strH;
#ifdef PROFILE
    HostFPrintF(HostFp,"initializeList(void) %lu\n",TimGetTicks());
#endif
    StnTrains = NULL;
    StnNameBuf = NULL;
    AllTrains = NULL;
    Trains = NULL;
    ListFormPtr = NULL;
    ListMenuBar = MenuInit(TableListMenu);
    TimeTableDB = NULL;
    strH = DmGetResource('tSTR',SeparatorStr);
    if (strH) {
	setSeparator(MemHandleLock(strH));
	MemHandleUnlock(strH);
	DmReleaseResource(strH);
    } else {
	setSeparator(" !\"#$%&()*+,-./:;<=>?@[\\]^_`{|}~");
    }
    initializeDB();
}

void initializeListAfterPrefs(void) {
    initializeDBAfterPrefs();
}

void terminateList(void) {
    closeList();
    MenuDispose(ListMenuBar);
    DmCloseDatabase(MemoDB);
    if (TimeTableDB) DmCloseDatabase(TimeTableDB);
}

/* Global variables can't be used in this function */
void jumpAlarmTriggered(SysAlarmTriggeredParamType *cmdPBP,
			ULong recID, LocalID dbID, UInt cardNo) {
    stepAlarmTriggered(cmdPBP,recID,dbID,cardNo);
}
