/*
 * TimeTable Viewer
 * Copyright(c)1999 by Hiroaki Sengoku <sengoku@gcd.org>
 * Version 0.1	Jun 19, 1999
 * Version 0.2	Jun 20, 1999
 * Version 0.3	Jun 20, 1999
 * Version 0.4	Jun 21, 1999
 * Version 0.5	Jun 22, 1999	AboutDialog
 * Version 0.6	Jun 25, 1999	SelectTrain, HardButton, PrefsDialog
 * Version 0.7	Jun 26, 1999	scroll bug fix
 * Version 0.8	Jul  3, 1999	Arrive, ShareReg, History, Transfer, StopClock
 * Version 0.9	Jul  8,	1999	multi '>', ContFile, CommentMark, scramble
 * Version 1.0	Jul 31, 1999	split bar, DB, ;;CANCEL, '\' esc, HeadType, LRU
 * Version 1.1	Aug 26, 1999	checkTimeOut
 * Version 1.2	Aug 27, 2000	Alarm, InstalledSec, LastUsedSec
 * Version 1.3  Aug  1, 2001	GPL version
 */
#define TTV_MAIN
#include "Common.h"

/* Table.c */
extern Boolean MainFormHandleEvent(EventPtr e);
extern void initializeTable(void);
extern void initializeTableDefault(void);
extern void initializeTableAfterPrefs(void);
extern void terminateTable(void);
extern void terminateTableBeforePrefs(void);
extern FileMenuType *getFileMenuItem(ULong now, short item);
extern short StopClockInterval;
extern short ArrivesColumn;
extern FreqType LargeClockFreq;
extern Boolean ChooseStartupFlag;
extern Boolean UnderlineSpecialFlag;
extern Boolean ScrollByKeyFlag;
extern Boolean OmitHourFlag;
extern short NarrowModeDefault;
extern Boolean TwelveFlag;
extern FileMenuType FileMenu[];
extern short BorderMin;
extern HeadType ArriveTimeHeadType;
extern short SelLine;
extern short SelItem;
extern short OfsArrive;
extern short NLines;
extern Boolean ManualSelectFlag;
extern Boolean LargeClockFlag;
extern Boolean ShowCommentsFlag;
extern Boolean ShowArrivesFlag;
extern long ConfigRecordNum;
extern long TrainTimeRecordNum;
extern long History[];
extern short HistoryTop;
extern short HistoryCur;
extern short HistoryBot;
extern HisAttrType HisAttr[];
extern short LRU_Top;
extern LRU_Type LRU[];
extern short TransNLines;
extern short TimeX;
extern short TimeY;

/* List.c */
extern Boolean TableListFormHandleEvent(EventPtr e);
extern void initializeList(void);
extern void initializeListAfterPrefs(void);
extern void terminateList(void);
extern void jumpAlarmTriggered(SysAlarmTriggeredParamType *cmdPBP,
			       ULong recID, LocalID dbID, UInt cardNo);
extern short SelStn;
extern short StnTop;
extern short StnSepWidth;
extern short ListTop;
extern ListOrderType ListOrder;

/* DB.c */
extern DmOpenRef TimeTableDB;
extern VoidHand TimeTableAppInfoH;
extern UInt TimeTableCategory;
extern WeekType Today;
extern ULong TodaySetSeconds;
extern InsertPosType InsertPosDB;
extern InsertPosType InsertPosMemo;

/* Dialog.c */
extern Word MidiIndex;
extern short MinBefore;
extern Byte SeparatorMap[];


/* preferences */

#define TimeTablePrefsCategoryName	"Prefs"
#define TimeTableSaveVersion		8	/* < TimeTableUnsaveVersion */
typedef struct {
    short version;		/* big endian is assumed */
    short HardButtonOverride;			/* save_ver >= 1 */
    short StopClockInterval;
    short ArrivesColumn;
    FreqType LargeClockFreq;
    Boolean ChooseStartupFlag;
    Boolean UnderlineSpecialFlag;
    Boolean ScrollByKeyFlag;
    Boolean OmitHourFlag;
    Boolean obsolete_1;
    char obsolete_2[STRMAX+1];
    FileMenuType FileMenu[MENU_NITEMS];		/* save_ver >= 2 */
    HeadType ArriveTimeHeadType;
    ListOrderType ListOrder;			/* save_ver >= 3 */
    InsertPosType InsertPosDB;
    InsertPosType InsertPosMemo;
    Word MidiIndex;				/* save_ver >= 4 */
    short MinBefore;
    short BorderMin;				/* save_ver >= 5 */
    FileMenuType FileMenuPM[MENU_NITEMS];
    Byte SeparatorMap[SeparatorMapLen];		/* save_ver >= 6 */
    short NarrowModeDefault;			/* save_ver >= 7 */
    Boolean TwelveFlag;				/* save_ver >= 8 */
} TimeTableSaveType;

#define TimeTableUnsaveVersion		17	/* > TimeTableSaveVersion */
typedef struct {
    short version;		/* big endian is assumed */
    short SelLine;
    short SelItem;
    short OfsArrive;
    short NLines;
    short SelStn;
    short StnTop;
    short StnSepWidth;
    short ListTop;
    Boolean ManualSelectFlag;
    Boolean LargeClockFlag;
    Boolean ShowCommentsFlag;
    Boolean ShowArrivesFlag;
    WeekType Today;
    ULong TodaySetSeconds;
    ULong timSeconds;
    long ConfigRecordNum;
    long TrainTimeRecordNum;
    long History[HISTORYMAX];
    short HistoryTop;
    short HistoryCur;
    short HistoryBot;
    HisAttrType HisAttr[HISTORYMAX];
    short LRU_Top;
    LRU_Type LRU[LRU_MAX];
} TimeTableUnsaveType;

char *PrivErrExtraStr;
short HardButtonOverride;
MenuBarPtr CurrentMenuBar;
ULong LastSec;
static UInt PrefsCategory;


/* error handling */

static char *PrivErrString(PrivErrType err) {
    switch (err) {
    case privErrNotEnoughSpace:
	return "Out of memory.";
    case privErrCanNotHappen:
	return "This can't happen.";
    default:
	return " ";
    }
}

static void PrivErrDisplay(char *str, Boolean fatal) {
    Word ret;
    if (PrivErrExtraStr == NULL) PrivErrExtraStr = " ";
    ret = FrmCustomAlert(PrivErrAlert,str,(fatal ?
			 "  Switch to other applications immediately.  "
		       : "  "),PrivErrExtraStr);
    PrivErrExtraStr = NULL;
}

static void PrivErrSetCode(Long err) {
    static char str[15];
    StrCopy(str,"ID=");
    StrIToA(str+3,err);
    PrivErrExtraStr = str;
}

/* clock */

static void digit2str(Byte x, char *p, Boolean zero) {
    short i;
    if (x > 100) x %= 100;
    i = x/10;
    if (i || zero) {
	*p++ = x/10 + '0';
    } else {
	*p++ = ' ';
    }
    *p++ = x%10 + '0';
    *p = '\0';
}

static void drawClock(void) {
    static ULong prevTime;
    ULong time = TimGetSeconds();
    DateTimeType dateTime;
    char str[10];
    WinHandle win = WinGetDrawWindow();
    if (time == prevTime) return;
    prevTime = time;
    WinSetDrawWindow(WinGetDisplayWindow());
    TimSecondsToDateTime(time,&dateTime);
    digit2str(HOUR(dateTime.hour),str,true);
    str[2] = ':';
    digit2str(dateTime.minute,&str[3],true);
    str[5] = ':';
    digit2str(dateTime.second,&str[6],true);
    WinDrawChars(str,StrLen(str),TimeX,TimeY);
    WinSetDrawWindow(win);
}

/* misc */

static void setToday(Boolean manual) {
    DateTimeType dateTime;
    ULong now = TimGetSeconds();
    ULong diff;
    short hour;
    diff = now - TodaySetSeconds;
    if (diff < 10 * 60) return;		/* 10 min. */
    TimSecondsToDateTime(now,&dateTime);
    hour = dateTime.hour;
    if (hour <= MidNight) {	/* till 4:00 */
	hour += 24;
	if (dateTime.weekDay == 0) dateTime.weekDay = 6;
	else dateTime.weekDay--;
    }
    if (Today == Holiday) {
	if (hour >= diff/(60*60) + MidNight) return;
    }
    if (dateTime.weekDay == 0) Today = Sunday;
    else Today = Monday + (dateTime.weekDay - 1);
    if (manual) TodaySetSeconds = now;
}

/* event handling */

Boolean menuAppear(void) {
    if (CurrentMenuBar) {
	return CurrentMenuBar->attr.visible;
    }
    return false;
}

static Boolean ApplicationHandleEvent(EventPtr e) {
    FormPtr frm;
    Word formId;
    Boolean handled = false;
    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);
	switch(formId) {
	case MainForm:
	    FrmSetEventHandler(frm,MainFormHandleEvent);
	    break;
	case TableListForm:
	    FrmSetEventHandler(frm,TableListFormHandleEvent);
	    break;
	case AlarmDialog:
	    FrmSetEventHandler(frm,AlarmDialogHandleEventP);
	    break;
	}
	handled = true;
    }
    return handled;
}

static Boolean preHandleEvent(EventPtr e) {
    ULong sec = TimGetSeconds();
    EventType event;
    Word err;
    Boolean handled = false;
    Word item;
    FileMenuType *menu;
    if (e->eType == keyDownEvent
	&& hard1Chr <= e->data.keyDown.chr
	&& e->data.keyDown.chr <= hard4Chr
	&& !(e->data.keyDown.modifiers & poweredOnKeyMask)
	&& (HardButtonOverride < 0 || sec - LastSec < HardButtonOverride)) {
	if (menuAppear()) {
	    event.eType = keyDownEvent;
	    event.data.keyDown.chr = menuChr;
	    event.data.keyDown.modifiers = 0;
	    MenuHandleEvent(CurrentMenuBar, &event, &err);
	}
	item = (e->data.keyDown.chr - hard1Chr) * 2;
	menu = getFileMenuItem(sec,item);
	if (menu->recordNum == TrainTimeRecordNum) item++;
	event.eType = menuEvent;
	event.data.menu.itemID = FileMenuOne + item;
	handled = FrmDispatchEvent(&event);
    }
    return handled;
}

/* initialize */

/* Global variables can't be used in this function */
static VoidHand getAppInfo(DmOpenRef dbP) {
    Err error;
    UInt cardNo;
    LocalID dbID;
    LocalID appInfoID;
    error = DmOpenDatabaseInfo(dbP,&dbID,NULL,NULL,&cardNo,NULL);
    if (error) {
	FrmCustomAlert(PrivErrAlert,"Get getting app info block"," "," ");
	return NULL;
    }
    error = DmDatabaseInfo (cardNo,dbID,NULL,NULL,NULL,NULL,NULL,
			    NULL,NULL,&appInfoID,NULL,NULL,NULL);
    if (error) {
	FrmCustomAlert(PrivErrAlert,"Get getting app info block"," "," ");
	return NULL;
    }
    return (VoidHand)MemLocalIDToGlobal(appInfoID, cardNo);
}

static void loadDefault(void) {
    short i;
    ConfigRecordNum = -1;
    TrainTimeRecordNum = -1;
    LargeClockFlag = true;
    for (i=0; i < MENU_NITEMS*2; i++) {
	FileMenu[i].str[0] = '\0';
	FileMenu[i].recordNum = -1;
    }
    ShowCommentsFlag = false;
    SelLine = SelItem = -1;
    ManualSelectFlag = false;	/* automatically unselected */
    HardButtonOverride = 4;		/* 4 sec */
    LargeClockFreq = Mostly;
    UnderlineSpecialFlag = true;
    ScrollByKeyFlag = false;
    ShowArrivesFlag = false;
    OfsArrive = 0;
    NarrowModeDefault = -1;
    TwelveFlag = false;
    OmitHourFlag = false;
    ArrivesColumn = 4;
    ArriveTimeHeadType = HeadNone;
    StopClockInterval = 1;	/* 3 min */
    ChooseStartupFlag = false;
    ListOrder = ListOrder_DB_Memo;
    InsertPosDB = InsertPosBot;
    InsertPosMemo = InsertPosBot;
    NLines = -1;
    TransNLines = 0;
    TodaySetSeconds = 0;
    setToday(false);
    ListTop = 0;
    StnTop = 0;
    SelStn = 0;
    StnSepWidth = 0;
    MidiIndex = 0;
    MinBefore = 3;
    BorderMin = -3;	/* no border, both */
    HistoryTop = HistoryCur = HistoryBot = 0;
    for (i=0; i < HISTORYMAX; i++) {
	History[i] = -1;
	HisAttr[i].StopClock = 0;
	HisAttr[i].ManualSelectFlag = false;
	HisAttr[i].SelItem = -1;
	HisAttr[i].SelLine = -1;
    }
    initializeTableDefault();
}

static Err TimeTableAppInfoInit(DmOpenRef db) {
    UInt cardNo;
    VoidHand h;
    LocalID dbID, appInfoID;
    TimeTableAppInfoType *appInfo;
    if (DmOpenDatabaseInfo(db,&dbID,NULL,NULL,&cardNo,NULL))
	return dmErrInvalidParam;
    if (DmDatabaseInfo(cardNo,dbID,NULL,NULL,NULL,NULL,NULL,
		       NULL,NULL,&appInfoID,NULL,NULL,NULL))
	return dmErrInvalidParam;
    if (appInfoID == NULL) {
	h = DmNewHandle(db,sizeof(TimeTableAppInfoType));
	if (h == NULL) return dmErrMemError;
	appInfoID = MemHandleToLocalID(h);
	DmSetDatabaseInfo(cardNo,dbID,NULL,NULL,NULL,NULL,NULL,
			  NULL,NULL,&appInfoID,NULL,NULL,NULL);
    }
    appInfo = MemLocalIDToLockedPtr(appInfoID,cardNo);
    DmSet(appInfo,0,sizeof(TimeTableAppInfoType),0);
    CategoryInitialize((AppInfoPtr)appInfo,LocalizedAppInfoStr);
    MemPtrUnlock(appInfo);
    return 0;
}

static void initTTVDB(void) {
    DmOpenRef db;
    Err err;
#ifdef PROFILE
    HostFPrintF(HostFp,"initTTVDB(void) %lu\n",TimGetTicks());
#endif
    db = DmOpenDatabaseByTypeCreator('DATA',APPID,dmModeReadOnly);
    if (!db) {
	err = DmCreateDatabase(0,"TimeTable Viewer Prefs",APPID,'DATA',false);
	if (err) {
	    PrivErrSetCode(err);
	    PrivErrDisplay("Can't create Prefs DB",false);
	    return;
	}
	db = DmOpenDatabaseByTypeCreator('DATA',APPID,dmModeReadWrite);
	if (!db) {
	    PrivErrSetCode(DmGetLastErr());
	    PrivErrDisplay("Can't open Prefs DB",false);
	    return;
	}
	err = TimeTableAppInfoInit(db);
	if (err) {
	    PrivErrSetCode(err);
	    PrivErrDisplay("Can't init AppInfo",false);
	    goto error;
	}
    }
    TimeTableCategory = CategoryFind(db,"TimeTable");
    PrefsCategory = CategoryFind(db,TimeTablePrefsCategoryName);
    if (TimeTableCategory == dmAllCategories
	|| PrefsCategory == dmAllCategories) {
	PrivErrDisplay("Can't find category",false);
	goto error;
    }
    DmCloseDatabase(db);
    TimeTableDB = DmOpenDatabaseByTypeCreator('DATA',APPID,dmModeReadOnly);
    TimeTableAppInfoH = getAppInfo(TimeTableDB);
    return;
 error:
    DmCloseDatabase(db);
    TimeTableDB = NULL;
}

static void loadPrefs(void) {
    DmOpenRef prefsDB;
    UInt category;
    UInt index;
    VoidHand record;
    TimeTableSaveType save;
    TimeTableUnsaveType unsave;
    ULong now, timSeconds;
    short i;
    prefsDB = DmOpenDatabaseByTypeCreator('DATA',APPID,dmModeReadOnly);
    if (!prefsDB) return;
    category = CategoryFind(prefsDB,TimeTablePrefsCategoryName);
    if (category == dmAllCategories)
	PrivErrDisplay("Can't find category. Removed ?",false);
    index = 0;
    record = DmQueryNextInCategory(prefsDB,&index,category);
    if (!record) {
	PrivErrSetCode(DmGetLastErr());
	PrivErrDisplay("Can't find save-prefs, Removed ?",false);
	goto close;
    }
    MemMove(&save,MemHandleLock(record),MemHandleSize(record));
    MemHandleUnlock(record);
    switch (save.version) {
    case 8:
	TwelveFlag = save.TwelveFlag;
    case 7:
	NarrowModeDefault = save.NarrowModeDefault;
    case 6:
	for (i=0; i < SeparatorMapLen; i++) {
	    SeparatorMap[i] = save.SeparatorMap[i];
	}
    case 5:
	BorderMin = save.BorderMin;
	for (i=0; i < MENU_NITEMS; i++) {
	    FileMenu[MENU_NITEMS+i] = save.FileMenuPM[i];
	}
    case 4:
	MidiIndex = save.MidiIndex;
	MinBefore = save.MinBefore;
    case 3:
	ListOrder = save.ListOrder;
	InsertPosDB = save.InsertPosDB;
	InsertPosMemo = save.InsertPosMemo;
    case 2:
	ArriveTimeHeadType = save.ArriveTimeHeadType;
    case 1:
	for (i=0; i < MENU_NITEMS; i++) {
	    FileMenu[i] = save.FileMenu[i];
	}
	if (save.version <= 4) {
	    for (i=0; i < MENU_NITEMS; i++) {
		FileMenu[MENU_NITEMS+i] = FileMenu[i];
	    }
	}
	HardButtonOverride = save.HardButtonOverride;
	StopClockInterval = save.StopClockInterval;
	LargeClockFreq = save.LargeClockFreq;
	ArrivesColumn = save.ArrivesColumn;
	if (ArrivesColumn < 1 || 8 < ArrivesColumn) ArrivesColumn = 1;
	ChooseStartupFlag = save.ChooseStartupFlag;
	UnderlineSpecialFlag = save.UnderlineSpecialFlag;
	ScrollByKeyFlag = save.ScrollByKeyFlag;
	OmitHourFlag = save.OmitHourFlag;
	break;
    default:
	PrivErrDisplay("Unknown version of save-prefs, ignored",false);
    }
    index++;
    record = DmQueryNextInCategory(prefsDB,&index,category);
    if (!record) {
	PrivErrSetCode(DmGetLastErr());
	PrivErrDisplay("Can't find unsave-prefs, Removed ?",false);
	goto close;
    }
    MemMove(&unsave,MemHandleLock(record),MemHandleSize(record));
    MemHandleUnlock(record);
#ifdef DEBUG
    HostFPrintF(HostFp,"loadPrefsDB unsave: "
		"index %d, size %ld, unsave_ver %d\n",
		index,MemHandleSize(record),unsave.version);
#endif
    if (unsave.version == TimeTableUnsaveVersion) {
	SelLine = unsave.SelLine;
	SelItem = unsave.SelItem;
	ManualSelectFlag = unsave.ManualSelectFlag;
	OfsArrive = unsave.OfsArrive;
	NLines = unsave.NLines;
	SelStn = unsave.SelStn;
	StnSepWidth = unsave.StnSepWidth;
	StnTop = unsave.StnTop;
	ListTop = unsave.ListTop;
	LargeClockFlag = unsave.LargeClockFlag;
	ShowCommentsFlag = unsave.ShowCommentsFlag;
	ShowArrivesFlag = unsave.ShowArrivesFlag;
	timSeconds = unsave.timSeconds;
	Today = unsave.Today;
	TodaySetSeconds = unsave.TodaySetSeconds;
	now = TimGetSeconds();
	if (now - timSeconds >= 8 * 60 * 60) SelLine = SelItem = -1;
	ConfigRecordNum = unsave.ConfigRecordNum;
	TrainTimeRecordNum = unsave.TrainTimeRecordNum;
	for (i=0; i < HISTORYMAX; i++) History[i] = unsave.History[i];
	HistoryTop = unsave.HistoryTop;
	HistoryCur = unsave.HistoryCur;
	HistoryBot = unsave.HistoryBot;
	for (i=0; i < HISTORYMAX; i++) HisAttr[i] = unsave.HisAttr[i];
	for (i=0; i < LRU_MAX; i++) LRU[i] = unsave.LRU[i];
	LRU_Top = unsave.LRU_Top;
    }
 close:
    DmCloseDatabase(prefsDB);
}

static void savePrefs(void) {
    DmOpenRef db;
    Err err;
    VoidHand record;
    TimeTableSaveType save;
    TimeTableUnsaveType unsave;
    UInt index;
    Boolean newrecord = false;
    short i;
    save.version = TimeTableSaveVersion;
    for (i=0; i < MENU_NITEMS; i++) {
	save.FileMenu[i] = FileMenu[i];
	save.FileMenuPM[i] = FileMenu[MENU_NITEMS+i];
    }
    for (i=0; i < SeparatorMapLen; i++) {
	save.SeparatorMap[i] = SeparatorMap[i];
    }
    save.HardButtonOverride = HardButtonOverride;
    save.StopClockInterval = StopClockInterval;
    save.LargeClockFreq = LargeClockFreq;
    save.ArrivesColumn = ArrivesColumn;
    save.ChooseStartupFlag = ChooseStartupFlag;
    save.UnderlineSpecialFlag = UnderlineSpecialFlag;
    save.ScrollByKeyFlag = ScrollByKeyFlag;
    save.NarrowModeDefault = NarrowModeDefault;
    save.TwelveFlag = TwelveFlag;
    save.OmitHourFlag = OmitHourFlag;
    save.ArriveTimeHeadType = ArriveTimeHeadType;
    save.ListOrder = ListOrder;
    save.InsertPosDB = InsertPosDB;
    save.InsertPosMemo = InsertPosMemo;
    save.MidiIndex = MidiIndex;
    save.MinBefore = MinBefore;
    save.BorderMin = BorderMin;
    unsave.version = TimeTableUnsaveVersion;
    if (ManualSelectFlag) {
	unsave.SelLine = SelLine;
	unsave.SelItem = SelItem;
    } else {
	unsave.SelLine = -1;
	unsave.SelItem = -1;
    }
    unsave.ManualSelectFlag = ManualSelectFlag;
    unsave.OfsArrive = OfsArrive;
    unsave.NLines = NLines;
    unsave.SelStn = SelStn;
    unsave.StnSepWidth = StnSepWidth;
    unsave.StnTop = StnTop;
    unsave.ListTop = ListTop;
    unsave.LargeClockFlag = LargeClockFlag;
    unsave.ShowCommentsFlag = ShowCommentsFlag;
    unsave.ShowArrivesFlag = ShowArrivesFlag;
    unsave.Today = Today;
    unsave.TodaySetSeconds = TodaySetSeconds;
    unsave.timSeconds = TimGetSeconds();
    unsave.ConfigRecordNum = ConfigRecordNum;
    unsave.TrainTimeRecordNum = TrainTimeRecordNum;
    for (i=0; i < HISTORYMAX; i++) unsave.History[i] = History[i];
    unsave.HistoryTop = HistoryTop;
    unsave.HistoryCur = HistoryCur;
    unsave.HistoryBot = HistoryBot;
    for (i=0; i < HISTORYMAX; i++) unsave.HisAttr[i] = HisAttr[i];
    for (i=0; i < LRU_MAX; i++) unsave.LRU[i] = LRU[i];
    unsave.LRU_Top = LRU_Top;
    /* write to DB */
    db = DmOpenDatabaseByTypeCreator('DATA',APPID,dmModeReadWrite);
    if (!db) {
	PrivErrSetCode(DmGetLastErr());
	PrivErrDisplay("Can't open Prefs DB",false);
	return;
    }
    index = 0;
    record = DmQueryNextInCategory(db,&index,PrefsCategory);
    if (!record) {
	index = 0;
	record = DmNewRecord(db,&index,sizeof(TimeTableSaveType));
	newrecord = true;
    } else {
	record = DmGetRecord(db,index);
	if ((i=MemHandleSize(record)) != sizeof(TimeTableSaveType)) {
	    record = DmResizeRecord(db,index,sizeof(TimeTableSaveType));
	}
    }
    if (record) {
	err = DmWrite(MemHandleLock(record),0,&save,sizeof(TimeTableSaveType));
	if (err) {
	    PrivErrSetCode(err);
	    PrivErrDisplay("Can't write save-prefs",false);
	}
	MemHandleUnlock(record);
	DmReleaseRecord(db,index,true);
    } else {
	PrivErrSetCode(DmGetLastErr());
        PrivErrDisplay("Can't open save-prefs",false);
	goto close;
    }
    index++;
    record = DmQueryNextInCategory(db,&index,PrefsCategory);
    if (!record) {
	record = DmNewRecord(db,&index,sizeof(TimeTableUnsaveType));
	newrecord = true;
    } else {
	record = DmResizeRecord(db,index,sizeof(TimeTableUnsaveType));
    }
    if (record) {
	err = DmWrite(MemHandleLock(record),0,&unsave,
		      sizeof(TimeTableUnsaveType));
	if (err) {
	    PrivErrSetCode(err);
	    PrivErrDisplay("Can't write unsave-prefs",false);
	}
	MemHandleUnlock(record);
	DmReleaseRecord(db,index,true);
    } else {
	PrivErrSetCode(DmGetLastErr());
	PrivErrDisplay("Can't open unsave-prefs",false);
	goto close;
    }
    if (newrecord) {
	err = DmMoveCategory(db,PrefsCategory,dmUnfiledCategory,true);
	if (err) {
	    PrivErrSetCode(err);
	    PrivErrDisplay("Can't move category",false);
	}
    }
 close:
    DmCloseDatabase(db);
}

static void initialize(void) {
    PrivErrExtraStr = NULL;
    LastSec = 0;
    initializeTable();
    initializeList();
    loadDefault();
    loadPrefs();
    initializeTableAfterPrefs();
    initializeListAfterPrefs();
    initTTVDB();
}

/* Get preferences, open (or create) app database */
static Word StartApplication(void) {
#if defined(DEBUG) || defined(PROFILE)
    HostFp = HostFOpen("T:/src/palm/table/debug.log","a");
    HostFPrintF(HostFp,"StartApplication %lu\n",TimGetTicks());
#endif
    initialize();
    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void StopApplication(void) {
    terminateTableBeforePrefs();
    savePrefs();
    terminateTable();
    terminateList();
#ifdef DEBUG
    HostFClose(HostFp);
#endif
}

/* The main event loop */
static void EventLoop(void) {
    Word err;
    EventType e;
    Boolean handled;
#ifdef PROFILE
    HostFPrintF(HostFp,"EventLoop(void) %lu\n",TimGetTicks());
#endif
    do {
	EvtGetEvent(&e,10);	/* 0.1 sec */
	if (!preHandleEvent(&e) && !SysHandleEvent(&e)) {
	    handled = false;
	    if (CurrentMenuBar) {
		handled = MenuHandleEvent(CurrentMenuBar, &e, &err);
	    }
	    if (!handled)
		if (!ApplicationHandleEvent(&e))
		    handled = FrmDispatchEvent(&e);
	}
    } while (e.eType != appStopEvent);
}

/* Global variables can't be used in this function */
void alarmTriggered(SysAlarmTriggeredParamType *cmdPBP) {
    DmOpenRef db;
    VoidHand appInfoH;
    TimeTableAppInfoType *appInfoP;
    ULong alarmRecID;
    LocalID alarmDbID;
    UInt alarmCardNo;
    db = DmOpenDatabaseByTypeCreator('DATA',APPID,dmModeReadOnly);
    if (!db) goto error;
    appInfoH = getAppInfo(db);
    DmCloseDatabase(db);
    if (!appInfoH) goto error;
    appInfoP = MemHandleLock(appInfoH);
    alarmRecID = appInfoP->alarmRecID;
    alarmDbID = appInfoP->alarmDbID;
    alarmCardNo = appInfoP->alarmCardNo;
    MemHandleUnlock(appInfoH);
    jumpAlarmTriggered(cmdPBP,alarmRecID,alarmDbID,alarmCardNo);
    return;
 error:
    jumpAlarmTriggered(cmdPBP,0,NULL,0);
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
    Word err;
    if (cmd == sysAppLaunchCmdNormalLaunch) {
	err = StartApplication();
	if (err) return err;
	EventLoop();
	StopApplication();
    } else if (cmd == sysAppLaunchCmdInitDatabase) {
	TimeTableAppInfoInit(((SysAppLaunchCmdInitDatabaseType*)cmdPBP)->dbP);
    } else if (cmd == sysAppLaunchCmdAlarmTriggered) {
	alarmTriggered((SysAlarmTriggeredParamType *)cmdPBP);
    } else {
	return sysErrParamErr;
    }
    return 0;
}
